# coding: utf-8
"""
    MXD.MappingDecomposer.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
from time import time
import numpy as np
from scipy.stats import linregress
from scipy.interpolate import UnivariateSpline
import molass_legacy.KekLib.DebugPlot as plt
from sklearn.mixture import GaussianMixture
from sklearn.cluster import KMeans
from molass_legacy.Peaks.RobustPeaks import RobustPeaks
from molass_legacy.Peaks.ElutionModels import egh
from Prob.GaussianMixture import hist_to_source
from Prob.EghMixture import EghMixture
import OurStatsModels as sm

APPLY_PARAM_CONSTRAINTS = True
USE_WLS = True

class CustomMixture(EghMixture):
    def __init__(self, K):
        self.K = K
        self.anim_data = False

    def initialize(self, X, init_params):
        self.N = X.shape[0]
        self.X = X
        self.X_  = X.flatten()
        sorted_params = np.array(sorted(init_params, key=lambda x:x[1]))
        w = np.sum(sorted_params[:,0])
        self.pi = sorted_params[:,0]/w
        self.tR = sorted_params[:,1]
        self.sigma = sorted_params[:,2]
        self.tau = sorted_params[:,3]
        if APPLY_PARAM_CONSTRAINTS:
            self.min_sigma = self.sigma*0.5
            self.max_sigma = self.sigma*1.5
            self.min_tR = np.max([np.zeros(self.K), self.tR - self.sigma], axis=0)
            self.max_tR = self.tR + self.sigma

class MappingDecomposer:
    def __init__(self, uv_rp, xr_rp, num_components=None):
        self.uv_rp = uv_rp
        self.xr_rp = xr_rp
        self.uv_x = uv_rp.x
        self.uv_y = uv_rp.y
        self.xr_x = xr_rp.x
        self.xr_y = xr_rp.y
        self.num_components = num_components
        self.uv_sy = hist_to_source(self.uv_x, self.uv_y)
        self.xr_sy = hist_to_source(self.xr_x, self.xr_y)

    def decompose_separatley(self):
        from Prob.EghMixture import EghMixture
        from Prob.EghMixtureUtils import get_components

        num_components = self.num_components
        x0 = self.uv_x
        y0 = self.uv_y
        sy0 = self.uv_sy
        mm0 = EghMixture(num_components)
        mm0.fit(np.expand_dims(sy0,1))
        cy_list0, ty0 = get_components(x0, y0, mm0)

        x1 = self.xr_x
        y1 = self.xr_y
        sy1 = self.xr_sy
        mm1 = EghMixture(num_components)
        mm1.fit(np.expand_dims(sy1,1))
        cy_list1, ty1 = get_components(x1, y1, mm1)

        plot_mapped_states([
            [x0, y0, cy_list0, ty0],
            [x1, y1, cy_list1, ty1],
            ])

    def decompose(self):
        mapping_params, params_info, w_ = self.guess_initial_params()
        print("mapping_params=", mapping_params)
        x0 = self.uv_x
        y0 = self.uv_y
        x1 = self.xr_x
        y1 = self.xr_y

        if False:
            for i in range(1):
                components_list = []
                mapping_points = []
                for (x, y), params_list in zip([(x0, y0), (x1, y1)], params_info):
                    cy_list = []
                    ty = np.zeros(len(x))
                    for params in params_list:
                        h, tR, sigma, tau = params
                        mapping_points.append((tR-sigma, tR, tR+sigma))
                        cy = egh(x, h, tR, sigma, tau)
                        cy_list.append(cy)
                        ty += ty
                    components_list.append((x, y, cy_list, ty))
                plot_mapped_states(components_list)

        uv_cm = CustomMixture(self.num_components)
        xr_cm = CustomMixture(self.num_components)

        uv_cm.initialize(np.expand_dims(self.uv_sy,1), params_info[0])
        xr_cm.initialize(np.expand_dims(self.xr_sy,1), params_info[1])

        for step in range(201):
            uv_cm._e_step()
            xr_cm._e_step()
            uv_params = uv_cm._m_step(step)
            xr_params = xr_cm._m_step(step)
            mapping_params_, uv_params_, xr_params_ = self.map_adjust_params(mapping_params, uv_params, xr_params, weights=w_)
            components_list = []
            for (x, y), params, pi_list in zip([(x0, y0), (x1, y1)], [uv_params_, xr_params_], [uv_cm.pi, xr_cm.pi]):
                cy_list = []
                ty = np.zeros(len(x))
                for prm, pi in zip(params, pi_list):
                    cy = egh(x, pi, *prm)
                    cy_list.append(cy)
                    ty += cy
                components_list.append((x, y, cy_list, ty))
            if step % 100 == 0:
                plot_mapped_states(components_list, scale=True, step=step, mappinn_info=[mapping_params_, uv_params_, uv_cm.pi, xr_params_, xr_cm.pi])
            uv_cm.update_params(step, uv_params_)
            xr_cm.update_params(step, xr_params_)

    def guess_initial_params(self):
        params_info = []
        mapping_points = []
        for k, rp in enumerate([self.uv_rp, self.xr_rp]):
            params_list = []
            x = rp.x
            y = rp.y
            w = []
            for peak in rp.get_peaks():
                ls, pt, rs = peak
                print([k], ls, pt, rs)
                mapping_points.append(x[peak])
                lsw = x[pt] - x[ls]
                tau = 0
                if self.num_components == 5:
                    point_list = [pt - 2*(pt-ls), ls]
                elif self.num_components in [6, 7]:
                    point_list = [pt - 3*(pt-ls), ls, (ls+pt)//2]
                elif self.num_components in [8]:
                    point_list = [pt - 4*(pt-ls), pt - 3*(pt-ls), ls, (ls+pt)//2]
                else:
                    assert False
                for ls_ in point_list:
                    h = y[ls_]
                    tR = x[ls_]
                    sigma = lsw/3
                    params_list.append((h, tR, sigma, tau))
                    w.append(1)
                rsw = x[rs] - x[pt]
                h = y[pt]
                tR = x[pt]
                sigma = (lsw+rsw)/3.5
                params_list.append((h, tR, sigma, tau))
                w.append(30)
                if self.num_components < 7:
                    point_list = [rs, pt + 2*(rs-pt)]
                elif self.num_components >= 7:
                    point_list = [rs, pt + 2*(rs-pt), pt + int(2.5*(rs-pt))]
                else:
                    assert False

                for rs_ in point_list:
                    h = y[rs_]
                    tR = x[rs_]
                    sigma = rsw/3
                    params_list.append((h, tR, sigma, tau))
                    w.append(1)
            params_info.append(params_list)

        w_ = w if USE_WLS else None
        slope, intercept = linear_regression(mapping_points[1], mapping_points[0])
        return [slope, intercept], params_info, w_

    def map_adjust_params(self, mapping_params, uv_params, xr_params, weights):
        slope, intercept = linear_regression(xr_params[:,0], uv_params[:,0], weights=weights)
        uv_params_ = []
        xr_params_ = []
        for uv_p, xr_p in zip(uv_params, xr_params):
            tR = (uv_p[0] + slope*xr_p[0] + intercept)/2
            sigma = (uv_p[1] + slope*xr_p[1])/2
            tau = (uv_p[2] + slope*xr_p[2])/2
            uv_params_.append((tR, sigma, tau))

            tR = ((uv_p[0] - intercept)/slope + xr_p[0])/2
            sigma = (uv_p[1]/slope + xr_p[1])/2
            tau =  (uv_p[2]/slope + xr_p[2])/2
            xr_params_.append((tR, sigma, tau))
        return [slope, intercept], np.array(uv_params_), np.array(xr_params_)

def linear_regression(x, y, weights=None):
    if weights is None:
        slope, intercept, r_value, p_value, std_err = linregress(x, y)
    else:
        X = sm.add_constant(x)
        mod = sm.WLS(y, X, weights=weights)
        res = mod.fit()
        intercept, slope = res.params
    return slope, intercept

def plot_mapped_states(components_list, scale=False, step=None, mappinn_info=None):
    plt.push()
    fig, (ax1, ax2, ax3) = plt.subplots(ncols=3, figsize=(21,7))

    step_str = "" if step is None else " step %d" % step
    fig.suptitle("UV-Xray Decomposed Mapping Figures (Plan)%s" % step_str, fontsize=30)
    ax1.set_title("UV Elution", fontsize=20)
    ax2.set_title("Xray Elution", fontsize=20)
    ax3.set_title("Mapped Elution", fontsize=20)

    for ax, components, color in zip([ax1, ax2], components_list, ['C0', 'C1']):
        x, y, cy_list, ty = components
        if scale:
            s = np.max(y)/np.max(ty)
        else:
            s = 1

        ax.plot(x, y, color=color)
        ax.plot(x, s*ty, ':', color='red')
        for cy in cy_list:
            ax.plot(x, s*cy, ':')

    if mappinn_info is not None:
        mapping_params, uv_params, uv_pi, xr_params, xr_pi = mappinn_info
        slope, intercept = mapping_params
        x0, y0, cy_list0, ty0 = components_list[0]
        x1, y1, cy_list1, ty1 = components_list[1]
        ax3.plot(x1, y1, color='C1', label='Xray')
        spline = UnivariateSpline(x0, y0, s=0, ext=3)
        x_ = x1*slope + intercept
        y_ = spline(x_)
        scale = np.max(y1)/np.max(y_)
        my = y_*scale
        ax3.plot(x1, my, color='gray', alpha=0.2, label='simply mapped UV')
        smy = make_mapped_uv(x1, my, uv_params, uv_pi, xr_params, xr_pi)
        ax3.plot(x1, smy, ':', color='C0', label='scaling mapped UV')
        ax3.legend(fontsize=16, loc="upper left")

    fig.tight_layout()
    fig.subplots_adjust(top=0.85)
    plt.show()
    plt.pop()

def make_mapped_uv(x, mapped_y, uv_params, uv_pi, xr_params, xr_pi):
    w = np.zeros(len(mapped_y))
    for uv_p, xr_p, params in zip(uv_pi, xr_pi, xr_params):
        w += egh(x, xr_p/uv_p, *params)
    smy = mapped_y*w
    scale = np.max(mapped_y)/np.max(smy)
    return smy*scale

def spike_demo():
    x0 = np.arange(1000)
    y0 = np.zeros(len(x0))
    for params in [
            [1.0, 401, 50, 5],
            [0.7, 602, 60, 10],
            ]:
        y0 += egh(x0, *params)

    x1 = np.arange(500)
    y1 = np.zeros(len(x1))
    for params in [
            # [0.6, 200, 40, 10],
            [0.6, 200, 30, 5],
            [1.0, 300, 30, 10],
            ]:
        y1 += egh(x1, *params)

    spike_demo_impl(x0, y0, x1, y1, num_components=2)

def spike_demo_impl(x0, y0, x1, y1, num_components=None):
    rb0 = RobustPeaks(x0, y0)
    rb1 = RobustPeaks(x1, y1)
    md = MappingDecomposer(rb0, rb1, num_components=num_components)
    # md.decompose_separatley()
    md.decompose()


def spike_demo_real(in_folder):
    from molass_legacy.Batch.StandardProcedure import StandardProcedure
    sp = StandardProcedure()
    sp.load(in_folder, debug=False)
    sd = sp.get_sd()
    uv_curve = sd.get_uv_curve()
    xr_curve = sd.get_xray_curve()
    spike_demo_impl(uv_curve.x, uv_curve.y, xr_curve.x, xr_curve.y, num_components=6)



# coding: utf-8
"""
    PreviewController.py

    Copyright (c) 2018-2021, SAXS Team, KEK-PF
"""
import logging
from molass_legacy.PeaksetSelector import PeakSetSelector
from ExtrapolationSolver import ExtrapolationSolver, run_extrapolation_solver
from ExtrapolSolverDialog import ExtrapolSolverDialog
from molass_legacy._MOLASS.SerialSettings import get_setting

class PreviewController:
    def __init__(self, dialog=None, editor=None):
        self.logger = logging.getLogger(__name__)
        self.parent_dialog = dialog
        self.editor = editor
        self.solver_results = None

    def run_solver( self, parent, pdata, popts,
                    known_info_list=None,
                    weighted=True, use_gui=True ):

        self.parent = parent
        self.mapper = mapper = pdata.mapper
        self.pdata = pdata
        self.popts = popts
        self.logger.info('cnv_ranges=%s, num_ranges=%s', str(pdata.cnv_ranges), str(pdata.num_ranges))

        self.arg_paired_ranges = pdata.paired_ranges    # keep this to call resursively in ExtrapolSolverDialog.solve_unknowns

        self.solver = ExtrapolationSolver(pdata, popts)

        self.selector = PeakSetSelector(pdata.cnv_ranges, mapper.x_curve)
        if known_info_list is None:
            known_info_list = get_setting('known_info_list')
        if known_info_list is not None:
            self.selector.update_known_peak_info_list(known_info_list)

        # parent_ = parent if aq_smoothness else None
        parent_ = parent

        ret_tuple = run_extrapolation_solver(
            self.solver, self.selector, parent=parent_, use_gui=use_gui,
            )
        self.solver_results = ret_tuple[0]
        self.peak_range_infos = ret_tuple[1]

        # assert use_elution_models
        self.use_elution_models = popts.use_elution_models

    def ok( self ):
        return self.solver_results is not None

    def show_dialog( self ):
        self.dialog = ExtrapolSolverDialog( self.parent_dialog.parent, self )
        self.dialog.show()

    def get_dialog( self ):
        return self.dialog

    def get_solver_ret_tuple( self ):
        return self.solver_results, self.peak_range_infos

    def get_better_results(self):
        # for compaitbility with LrfResultPool
        return self.solver_results



import sys
import os
import numpy    as np
from mpl_toolkits.mplot3d   import Axes3D
import matplotlib.pyplot    as plt
# import statsmodels.api      as sm

sys.path.append( os.path.dirname( os.path.abspath( __file__ ) ) + '/../../lib' )
import molass_legacy.KekLib, AutorgKek, SerialAnalyzer
import OurStatsModels       as sm
from SerialData             import SerialData
from molass_legacy.KekLib.NumpyUtils             import np_savetxt

sys.path.append( os.path.dirname( os.path.abspath( __file__ ) ) + '/../../..' )
from TestEnv import env_dict

data_folder = env_dict[ 'Dec05' ]
conc_folder = env_dict[ 'Dec05' ]

serial = SerialData( conc_folder, data_folder,  )
serial.wait_until_ready()
serial.make_adjusted_array( 5.0 )

print( 'serial.intensity_array.shape=', serial.intensity_array.shape )
print( 'serial.conc_array.shape=', serial.conc_array.shape )

x = serial.qvector
x_slice = slice( 0, 200 )
x_ = x[x_slice]
indeces = np.arange( 206, 233 )

x_i = 50
x_f = x_[x_i]

fig = plt.figure( figsize=(18, 9) )
ax1 = fig.add_subplot( 231 )
ax2 = fig.add_subplot( 232, projection='3d' )
ax3 = fig.add_subplot( 233 )
ax4 = fig.add_subplot( 234, projection='3d' )
ax5 = fig.add_subplot( 235 )
ax6 = fig.add_subplot( 236 )

ax1.set_title( 'Concentration Curve' )
ax1.set_xlabel( 'Seq No.' )
ax1.set_ylabel( 'Concentration' )

ax2.set_title( 'Intensity Curves on Q[0:200] and the plane at Q[50]=%.3g' % x_f )
ax2.set_xlabel( 'Q' )
ax2.set_ylabel( 'Seq No.' )
ax2.set_zlabel( 'Intensity' )

ax3.set_title( 'Intensity Variation with errorbar at Q[50]=%.3g' % x_f )
ax3.set_xlabel( 'Seq No.' )
ax3.set_ylabel( 'Concentration' )

ax4.set_title( 'Mutivariate Regression at Q[50]=%.3g' % x_f )
ax4.set_xlabel( 'C' )
ax4.set_ylabel( '-C**2' )
ax4.set_zlabel( 'Intensity' )

ax5.set_title( 'Extrapolated A(q)' )
ax5.set_xlabel( 'Q' )
ax5.set_ylabel( 'Intensity / C' )

ax6.set_title( 'Extrapolated B(q)' )
ax6.set_xlabel( 'Q' )
ax6.set_ylabel( 'Intensity / C**2' )

mc_vector = serial.mc_vector[indeces]
ax1.plot( indeces, mc_vector )

for i in indeces:
    c = serial.mc_vector[i]
    # y_ = np.ones( len(x_) ) * c
    y_ = np.ones( len(x_) ) * i
    z_ = serial.intensity_array[i,x_slice,1]
    ax2.plot( x_, y_, z_ )

ymin, ymax = ax2.get_ylim()
ax2.set_ylim( ymin, ymax )
zmin, zmax = ax2.get_zlim()
ax2.set_zlim( zmin, zmax )
ax2.plot( np.ones(5)*x_f,
            [ymin, ymax, ymax, ymin, ymin ],
            [zmin, zmin, zmax, zmax, zmin ],
            ':', color='black' )
ax2.plot( np.ones(len(mc_vector))*x_f, indeces, serial.intensity_array[indeces,x_i,1], ':', color='red' )

y_f = serial.intensity_array[indeces,x_i,1]
y_e = serial.intensity_array[indeces,x_i,2]
ax3.plot( indeces, y_f, ':', color='red' )
ax3.errorbar( indeces, y_f, yerr=y_e, color='blue', fmt='none' )

x1_ = mc_vector
x2_ = -mc_vector**2
X = np.array( [ x1_, x2_] ).T

param_list = []
error_list = []
for i in range(0,200):
    z_f = serial.intensity_array[indeces,i,1]
    e2  = serial.intensity_array[indeces,i,2]**2
    w   = 1/e2
    model   = sm.WLS(z_f, X, weights=w)
    result  = model.fit()
    param_list.append( result.params )
    # error_list.append( np.diag( result.cov_params() ) )
    error = np.sqrt( np.dot( model.XtWX_inv_XtW**2, e2 ) )
    error_list.append( error )

params = param_list[x_i]

z_f = serial.intensity_array[indeces,x_i,1]
z_r = np.dot( X, params )

ax4.scatter( x1_, x2_, z_f )
ax4.scatter( x1_, x2_, z_r, color='orange' )

xmin, xmax = ax4.get_xlim()
ymin, ymax = ax4.get_ylim()
ax4.set_xlim( xmin, xmax )
ax4.set_ylim( ymin, ymax )

xx, yy = np.meshgrid( np.linspace(xmin, xmax, 10), np.linspace(ymin, ymax, 10) )
zz = xx * params[0] + yy * params[1]
ax4.plot_surface(xx, yy, zz, alpha=0.1 )

params_array = np.array( param_list )
error_array = np.array( error_list )
# np_savetxt( 'params_array.csv', params_array )
np_savetxt( 'error_array.csv', error_array )

A = params_array[:,0]
B = params_array[:,1]
ax5.plot( x_, A, color='red' )
ax6.plot( x_, B, color='red' )
ax5.errorbar( x_f, A[x_i], yerr=error_array[x_i,0], color='blue', fmt='none' )
ax6.errorbar( x_f, B[x_i], yerr=error_array[x_i,1], color='blue', fmt='none' )

for ax in [ax5, ax6]:
    ymin, ymax = ax.get_ylim()
    ax.set_ylim( ymin, ymax )
    ax.plot( [ x_f, x_f ], [ymin, ymax], ':', color='black' )

plt.tight_layout()
plt.show()



"""
    BasicOptimizer.py

    Copyright (c) 2021-2025, SAXS Team, KEK-PF
"""
import logging
import numpy as np
from bisect import bisect_right
from time import time
import numpy as np
from bisect import bisect_right
from molass_legacy.KekLib.ExceptionTracebacker import log_exception
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.Baseline.Constants import SLOPE_SCALE
from molass_legacy.DataStructure.SvdDenoise import get_denoised_data
from molass_legacy._MOLASS.SerialSettings import get_setting
from .StateSequence import save_opt_params
from .RgDiscreteness import compute_rg_discreteness
from .MwIntegrity import compute_mw_integrity_impl
from .FvSynthesizer import synthesize
from .OptLrfInfo import OptLrfInfo
from molass_legacy._MOLASS.Version import is_developing_version
from .ValidComponents import ValidComponents

USE_COLUMN_INTERP = True

from molass_legacy.GuinierTools.GuinierDeviation import USE_NORMALIZED_RMSD_FOR_RGCURVES

USE_NORMALIZED_RMSD = True
USE_FROBENIUS_XDIFFMAX = False
USE_FEATURE_DEVIATION = False
USE_JSD = False
if USE_NORMALIZED_RMSD:
    from molass_legacy.Distance.NormalizedRmsd import normalized_rmsd
elif USE_FROBENIUS_XDIFFMAX:
    from molass_legacy.Distance.FrobeniusXdiffmax import frobenius_xdiffmax
elif USE_FEATURE_DEVIATION:
    from molass_legacy.Distance.DeviationMeasure import feature_deviation
elif USE_JSD:
    from molass_legacy.Distance.JensenShannon import deformed_jsd

USE_BOUNDS = True
AVOID_VANISHING_RATIO = 0.02    # minimum scale ratio against the max scale

PENALTY_SCALE = 1e3
UV_XR_RATIO_ALLOW = 0.5
UV_XR_RATIO_SCALE = 100
WEAK_PENALTY_SCALE = 0.01
SUPERIOR_2D_LRF_ALLOW = 0.1
SLOPE_ALLOWANCE = 0.05
INTERCEPT_ALLOWANCE = 0.05
SCALE_MAX_RATIO = 1.5
PARAMS_SCALE = 10
IMMEDIATE_KNOWN_BEST = False
BAD_PARAMS_RETURN = 1e8
SCORE_DEV_ALLOW = 0.25
SCORE_DEV_LIMIT = 0.5
MEAN_WEIGHT = 0.2
MAX_WEIGHT = 0.6
DEV_WEIGHT = 0.2
EVAL_PEAK_DEVIATION = False
PRIMARY_PEAK_DEV_ALLOW = 0.01
KRATKY_SMOOTHNESS_BAD_VALUE = 1.0
USE_COMPOSED_XR_COMPONENTS = False
USE_COMPOSED_UV_COMPONENTS = False
COERCE_BOUNDED_BQ = True
USE_RGCURVE_DEVIATION = True
ADJUST_2D_TARGET = 1

class BasicOptimizer:
    def __init__(self, dsets, n_components, params_type, kwargs):
        self.NUM_MAJOR_SCORES = get_setting("NUM_MAJOR_SCORES")

        self.devel = is_developing_version()
        self.logger = logging.getLogger(__name__)

        self.dsets = dsets
        self.n_components = n_components
        self.params_type = params_type
        self.use_K = params_type.use_K
        if self.use_K:
            self.nj, self.mej, self.tj, self.mpj = params_type.get_trans_indeces()

        for_split_only = kwargs.pop("for_split_only", False)
        if for_split_only:
            # as used in test_6690_BasinHopping.py
            return

        self.uv_base_curve = kwargs.pop("uv_base_curve", None)
        self.xr_base_curve = kwargs.pop("xr_base_curve", None)
        self.qvector = kwargs.pop("qvector", None)
        self.qzeros = np.zeros(len(self.qvector))
        self.wvector = kwargs.pop("wvector", None)  # lvector in sd
        self.uv_index = bisect_right(self.wvector, get_setting("absorbance_picking"))
        self.xr_index = bisect_right(self.qvector, get_setting("intensity_picking"))
        self.shm = kwargs.pop("shared_memory", None)
        if False:
            import inspect
            for frm in inspect.stack()[1:]:
                self.logger.info("optimizer: %s %s (%d)", frm.filename, frm.function, frm.lineno)

        self.logger.info("optimizer has been created with params_type=%s", str(self.params_type))

        self.separate_eoii = get_setting("separate_eoii")
        self.separate_eoii_type = get_setting("separate_eoii_type")
        self.separate_eoii_flags = get_setting("separate_eoii_flags")
        self.num_pure_components = self.n_components - (2 if self.separate_eoii else 1)

        composite = kwargs.pop("composite", None)
        if composite is None:
            from .CompositeInfo import CompositeInfo
            composite = CompositeInfo(n_components)
        self.composite = composite

        extra_num_components = self.composite.get_extra_num_components()
        self.num_pure_components = self.n_components - 1 - extra_num_components

        (xr_curve, xrD), rg_curve, (uv_curve, uvD) = dsets

        if USE_JSD:
            # negetive values are not allowed in scipy.spatial.distance.jensenshannon
            for curve in xr_curve, uv_curve:
                curve.y[curve.y < 0] = 0

        self.xr_curve = xr_curve
        self.xrD = xrD
        self.xrE = dsets.E
        self.xr_zero_scatter = np.zeros(xrD.shape[0])
        xr_rank  = self.n_components
        if self.separate_eoii_type == 1:
            xr_rank += 1
        elif self.separate_eoii_type == 2:
            # not yet supported
            xr_rank += np.sum(self.separate_eoii_flags, dtype=int)
        self.xrD_ = get_denoised_data(xrD, rank=xr_rank)
        self.xr_norm1 = np.linalg.norm(xr_curve.y)
        self.xr_norm2 = np.linalg.norm(xrD)
        self.rg_curve = rg_curve
        self.uv_curve = uv_curve
        self.uvD = uvD
        self.uvD_ = get_denoised_data(uvD, rank=self.n_components)

        if False:
            from MatrixData import simple_plot_3d
            plt.push()
            fig = plt.figure(figsize=(14,7))
            ax1 = fig.add_subplot(121, projection='3d')
            ax2 = fig.add_subplot(122, projection='3d')
            simple_plot_3d(ax1, uvD)
            simple_plot_3d(ax2, xrD)
            fig.tight_layout()
            plt.show()
            plt.pop()

        uv_scale = uv_curve.max_y / xr_curve.max_y
        self.uv_norm1 = self.xr_norm1 * uv_scale
        self.uv_norm2 = self.xr_norm2 * uv_scale
        self.uv_i = np.arange(uvD.shape[0])
        self.uv_j = uv_curve.x

        if USE_NORMALIZED_RMSD or USE_FROBENIUS_XDIFFMAX:
            self.uv2d_adjust = ADJUST_2D_TARGET - np.log10(self.uv_norm1)
            self.xr2d_adjust = ADJUST_2D_TARGET - np.log10(self.xr_norm1)

        if USE_NORMALIZED_RMSD_FOR_RGCURVES:
            self.rgcurve_adjust = -1.2
        else:
            self.rgcurve_adjust = -2.7

        if USE_COLUMN_INTERP:
            if False:
                import Optimizer.ColumnInterp
                reload(Optimizer.ColumnInterp)
            from .ColumnInterp import ColumnInterp
            self.uv_interp = ColumnInterp(self.uvD_)
        else:
            from scipy.interpolate import RectBivariateSpline
            self.uv_interp = RectBivariateSpline(self.uv_i, self.uv_j, self.uvD_, s=0)
        self.uv_shape = (uvD.shape[0], xrD.shape[1])
        self.given_seeds = None
        self.real_bounds = None
        self.eval_counter = 0
        self.callback_counter = 0
        self.svd_error_count = 0
        self.minima_props = None
        self.valid_components = np.ones(self.num_pure_components, dtype=bool)
        self.ratio_interpretation = get_setting("ratio_interpretation")
        self.poresize_bounds = get_setting("poresize_bounds")
        self.apply_rg_discreteness = get_setting("apply_rg_discreteness")
        self.apply_mw_integrity = get_setting("apply_mw_integrity")
        self.apply_discreteness = self.apply_rg_discreteness or self.apply_mw_integrity
        self.mw_integer_ratios = get_setting("mw_integer_ratios")
        self.c_indeces = None
        self.rg_discreteness_unit = get_setting("rg_discreteness_unit")
        self.kratky_smoothness = get_setting("kratky_smoothness")
        self.avoid_peak_fronting = get_setting("avoid_peak_fronting")
        self.apply_sf_bounds = get_setting("apply_sf_bounds")
        self.logger.info("ratio_interpretation=%d, poresize_bounds=%s, separate_eoii_type=%d, apply_rg_discreteness=%d, rg_discreteness_unit=%g",
            self.ratio_interpretation, str(self.poresize_bounds), self.separate_eoii_type, self.apply_rg_discreteness, self.rg_discreteness_unit)

        if self.apply_mw_integrity:
            self.logger.info("applying mw discreteness with ratios %s", str(self.mw_integer_ratios))

        self.exception_count = 0
        self.isnan_logged = False
        self.debug_fv = False

        # note that these weights are independent of composites variation
        self.uw, self.vw, self.W_ = dsets.get_opt_weight_info()
        self.xrDw = self.W_ * self.xrD_
        # self.rg_weights = self.rg_curve.get_weights()
        self.exports_bounds = False
        self.vc = ValidComponents(self.num_pure_components)

    def get_model_name(self):
        return self.params_type.get_model_name()

    def get_num_components(self):
        return self.num_pure_components

    def get_paramslider_info(self):
        return self.params_type.get_paramslider_info()

    def set_composite(self, composite):
        self.logger.info("this composite has been renewed to %s", str(composite))
        self.composite = composite

    def get_required_params(self, params):
        from molass_legacy._MOLASS.SerialSettings import set_setting
        num_params = self.params_type.num_params
        set_setting("init_sec_params", params[num_params:])
        return params[0:num_params], params[num_params:]

    def split_params_simple(self, p):
        self.separate_params = self.params_type.split_params_simple(p)
        return self.separate_params

    def split_as_unified_params(self, p, **kwargs):
        return self.params_type.split_as_unified_params(p, **kwargs)

    def get_parameter_names(self):
        return self.params_type.get_parameter_names()

    def solve(self, init_params, real_bounds=None, niter=100, seed=None, callback=True, method=None, debug=False, show_history=False):
        t0 = time()
        self.logger.info("solve started with niter=%d, seed=%s", niter, str(seed))
        self.logger.info("init_params=%s", str(init_params))

        self.prepare_for_optimization(init_params, real_bounds=real_bounds)

        self.fv_array = np.zeros(niter + 3)
        self.fv_array_size = 0
        self.min_fv = None
        self.min_i = None

        self.debug_fv = debug
        if callback:
            self.cb_fh = open("callback.txt", "w")
            self.write_init_callback_txt(init_params)

        if method is None:
            from .OptimizerUtils import get_impl_method_name
            method = get_impl_method_name()

        bounds = np.array([(0, 10)]*len(init_params))
        norm_params = self.to_norm_params(init_params)

        if method == "bh":
            from importlib import reload
            import molass_legacy.Solvers.BH.SolverBH
            reload(molass_legacy.Solvers.BH.SolverBH)
            from molass_legacy.Solvers.BH.SolverBH import SolverBH
            bh = SolverBH(self)
            result = bh.minimize(self.objective_func_wrapper, norm_params, niter=niter, seed=seed, bounds=bounds, show_history=show_history)

        elif method == "ultranest":
            from importlib import reload
            import molass_legacy.Solvers.UltraNest.SolverUltraNest
            reload(molass_legacy.Solvers.UltraNest.SolverUltraNest)
            from molass_legacy.Solvers.UltraNest.SolverUltraNest import SolverUltraNest
            ultranest = SolverUltraNest(self)
            result = ultranest.minimize(self.objective_func_wrapper, norm_params, niter=niter, seed=seed, bounds=bounds)
 
        elif method == "emcee":
            from importlib import reload
            import molass_legacy.Solvers.MCMC.SolverEmcee
            reload(molass_legacy.Solvers.MCMC.SolverEmcee)
            from molass_legacy.Solvers.MCMC.SolverEmcee import SolverEmcee
            mcmc = SolverEmcee(self)
            result = mcmc.minimize(self.objective_func_wrapper, norm_params, niter=niter, seed=seed, bounds=bounds)
 
        elif method == "pyabc":
            from importlib import reload
            import molass_legacy.Solvers.ABC.SolverPyABC
            reload(molass_legacy.Solvers.ABC.SolverPyABC)
            from molass_legacy.Solvers.ABC.SolverPyABC import SolverPyABC
            abc = SolverPyABC(self)
            result = abc.minimize(self.objective_func_wrapper, norm_params, niter=niter, seed=seed, bounds=bounds)

        elif method == "pymc":
            from importlib import reload
            import molass_legacy.Solvers.SMC.SolverPyMC
            reload(molass_legacy.Solvers.SMC.SolverPyMC)
            from molass_legacy.Solvers.SMC.SolverPyMC import SolverPyMC, get_picklable_func
            smc = SolverPyMC(self)
            func = get_picklable_func(self)
            result = smc.minimize(func, norm_params, niter=niter, seed=seed, bounds=bounds)
 
        else:
            raise ValueError("Unknown method: %s" % method)

        if callback:
            self.cb_fh.close()

        self.logger.info("solve finished in %.3g minutes %.3g seconds with %d iterations and %d evaluations.",
                            *divmod(time()-t0, 60), result.nit, result.nfev)

        if self.svd_error_count > 0:
            self.logger.warning("There were %d SVD errors.", self.svd_error_count)

        return result

    def prepare_for_optimization(self, init_params, real_bounds=None, minimally=False):
        self.eval_counter = 0
        self.init_params = init_params
        self.init_separate_params = self.split_params_simple(init_params)
        self.separate_params = self.init_separate_params    # to ensure that self.separate_params always exists
        xr_params, xr_baseparams, rgs, mapping, uv_params, uv_baseparams, mappable_range = self.init_separate_params[0:7]
        self.init_rgs = rgs
        self.init_uv_params = uv_params
        self.init_uv_baseparams = uv_baseparams
        self.init_mapping = mapping
        self.init_mappable_range = mappable_range
        if real_bounds is None:
            self.real_bounds = np.array(self.get_param_bounds(init_params))
            from_arg = False
        else:
            # for SDM
            self.real_bounds = real_bounds
            from_arg= True
        self.logger.info("from_arg=%s, real bounds=%s", from_arg, str(self.real_bounds))
        self.set_params_scale(self.real_bounds)
        self.bounds_mask = self.params_type.make_bounds_mask()
        self.update_bounds(self.init_params)
        masked_init_params = self.init_params[self.bounds_mask]
        self.zero_bounds = np.zeros(masked_init_params.shape)
        self.sf_bounds = None   # referenced in objective_func, but is this ok?

        self.update_minima_props(init_params)

        self.ones_nc = np.ones(xr_params.shape[0])
        self.init_xr_params = xr_params
        self.init_xr_baseparams = xr_baseparams
        init_slope, init_intercept = xr_baseparams[0:2]     # how about r?
        self.init_slope = init_slope
        self.init_intercept = init_intercept
        self.slope_allowance = (init_slope*SLOPE_ALLOWANCE)**2
        self.intercept_allowance = (init_intercept*INTERCEPT_ALLOWANCE)**2
        self.slope_penalty_scale = 1/max(1e-16, self.slope_allowance)
        self.intercept_penalt_scale = 1/max(1e-16, self.intercept_allowance)

        self.zeros_rg = np.zeros(len(self.init_separate_params[2]) - 1)
        if EVAL_PEAK_DEVIATION:
            self.update_primary_peak()          # must be called before update_guinier_region()

        # estimate approximate peak_points used in the optimization for SEC conformance evaluation
        lrf_info = self.objective_func(init_params, return_lrf_info=True)
        self.peak_points = lrf_info.estimate_xr_peak_points()
        self.update_guinier_region()

        if minimally:
            return

        self.xm, self.ym, self.rg = self.rg_curve.get_valid_curves()
        self.mask = self.rg_curve.get_mask()
        self.ones_rg = np.ones(len(self.rg))

        if self.separate_eoii_type > 0 and self.apply_sf_bounds:
            if True:
                from importlib import reload
                import molass_legacy.Optimizer.StructureFactorBounds
                reload(molass_legacy.Optimizer.StructureFactorBounds)
            from .StructureFactorBounds import StructureFactorBounds
            from molass_legacy.Kratky.GuinierKratkyInfo import GuinierKratkyInfo
            # consier using GuinierDeviation as an aternative
            gk_info = GuinierKratkyInfo(self, init_params, lrf_info)
            self.sf_bounds = StructureFactorBounds(self.qvector, lrf_info, gk_info)

    def update_minima_props(self, params):
        lrf_info = self.objective_func(params, return_lrf_info=True)
        self.minima_props = lrf_info.get_xr_proportions()
        self.valid_components = self.vc.get_valid_vector()
        self.logger.info("valid_components=%s", self.valid_components)
        num_valid_components = len(np.where(self.valid_components)[0])
        if num_valid_components > 1:
            self.zeros_valid_rg = np.zeros(num_valid_components - 1)
        else:
            # self.zeros_valid_rg will not be used in the optimization
            pass

    def write_init_callback_txt(self, init_params):
        self.eval_counter = -1
        fv  = self.objective_func(init_params)
        x = self.to_norm_params(init_params)
        self.minima_callback(x, fv, False)

    def compute_LRF_matrices(self, x, y, xr_cy_list, xr_ty, uv_x, uv_y, uv_cy_list, uv_ty, debug=False):
        Cxr = self.composite.compute_C_matrix(y, xr_cy_list, eoii=True, ratio_interpretation=self.ratio_interpretation)
        Pxr = (self.xrDw @ np.linalg.pinv(Cxr*self.vw))/self.uw
        Cuv = self.composite.compute_C_matrix(uv_y, uv_cy_list, ratio_interpretation=self.ratio_interpretation)
        if USE_COLUMN_INTERP:
            mapped_UvD = self.uv_interp(uv_x)
        else:
            mapped_UvD = self.uv_interp(self.uv_i, uv_x).reshape(self.uv_shape)
        Puv = mapped_UvD @ np.linalg.pinv(Cuv)

        if debug:
            from molass_legacy.DataStructure.MatrixData import simple_plot_3d

            with plt.Dp():
                fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5), subplot_kw=dict(projection="3d"))
                fig.suptitle("compute_LRF_matrices debug (1)")
                ax1.set_title("uvD")
                ax2.set_title("mapped_UvD")
                simple_plot_3d(ax1, self.uvD)
                simple_plot_3d(ax2, mapped_UvD)
                fig.tight_layout()
                plt.show()

        if USE_COMPOSED_XR_COMPONENTS:
            scaled_xr_cy_array = (Pxr[self.xr_index,:] * Cxr.T).T
            xr_ty = np.sum(scaled_xr_cy_array, axis=0)
        else:
            scaled_xr_cy_array = Cxr

        if USE_COMPOSED_UV_COMPONENTS:
            scaled_uv_cy_array = (Puv[self.uv_index,:] * Cuv.T).T
            uv_ty = np.sum(scaled_uv_cy_array, axis=0)
        else:
            scaled_uv_cy_array = Cuv

        if debug:
            print("self.uvD.shape=", self.uvD.shape)
            print("mapped_UvD.shape=", mapped_UvD.shape)

            def plot_componnts(ax, x, y, n, cy_list, ty=None, data_color=None):
                computing_ty = ty is None
                if computing_ty:
                    ty = np.zeros(len(x))
                ax.plot(x, y, label="data", color=data_color)
                for k, cy in enumerate(cy_list):
                    ax.plot(x, cy, ":", label="component-%d" % k)
                    if computing_ty and k < n:
                        ty += cy
                ax.plot(x, ty, ":", color="red", label="total")
                ax.legend()

            n = self.n_components
            with plt.Dp():
                fig, axes = plt.subplots(ncols=2, nrows=2, figsize=(12,10))
                fig.suptitle("compute_LRF_matrices debug (2)")
                ax11 = axes[0,0]
                ax12 = axes[0,1]
                ax21 = axes[1,0]
                ax22 = axes[1,1]

                ax11.set_title("UV components")
                ax12.set_title("XR components")
                ax21.set_title("UV scaled components")
                ax22.set_title("XR scaled components")

                plot_componnts(ax11, uv_x, uv_y, n, uv_cy_list, data_color="blue")
                plot_componnts(ax12, x, y, n, xr_cy_list, data_color="orange")
                plot_componnts(ax21, uv_x, uv_y, n, scaled_uv_cy_array, uv_ty, data_color="blue")
                plot_componnts(ax22, x, y, n, scaled_xr_cy_array, xr_ty, data_color="orange")

                fig.tight_layout()
                plt.show()

        if COERCE_BOUNDED_BQ:
            if self.sf_bounds is not None:
                Pxr[:,-1] = self.sf_bounds.compute_bounded_bq(Pxr)

        return OptLrfInfo(Pxr, Cxr, Puv, Cuv, mapped_UvD,
                    self.qvector, self.xrD, self.xrE, x, y, xr_ty, scaled_xr_cy_array, uv_x, uv_y, uv_ty, scaled_uv_cy_array, self.composite)

    def create_lrf_info_for_debug(self, x, y, xr_ty, xr_cy_list, uv_x, uv_y, uv_ty, uv_cy_list):
        from importlib import reload
        import molass_legacy.Optimizer.OptLrfInfoDebug
        reload(molass_legacy.Optimizer.OptLrfInfoDebug)
        from molass_legacy.Optimizer.OptLrfInfoDebug import OptLrfInfoProxy
        return OptLrfInfoProxy(self.qvector, self.xrD, self.xrE, x, y, xr_ty, xr_cy_list, uv_x, uv_y, uv_ty, uv_cy_list, self.composite)

    def update_rg_params(self, lrf_info):
        lrf_info.update_optimizer(self)

    def compute_fv(self, lrf_info, xr_params, rg_params, seccol_params, penalties, p, debug=False):
        Pxr, Cxr, Puv, Cuv, mapped_UvD = lrf_info.matrices

        if USE_NORMALIZED_RMSD:
            XR_2D_fitting = normalized_rmsd(lrf_info.xr_ty, lrf_info.y, adjust=self.xr2d_adjust)
            UV_2D_fitting = normalized_rmsd(lrf_info.uv_ty, lrf_info.uv_y, adjust=self.uv2d_adjust)
        elif USE_FROBENIUS_XDIFFMAX:
            XR_2D_fitting = frobenius_xdiffmax(lrf_info.xr_ty, lrf_info.y, adjust=self.xr2d_adjust)
            UV_2D_fitting = frobenius_xdiffmax(lrf_info.uv_ty, lrf_info.uv_y, adjust=self.uv2d_adjust)            
        elif USE_FEATURE_DEVIATION:
            XR_2D_fitting = feature_deviation(lrf_info.xr_ty, lrf_info.y, self.xr_curve.max_y, self.xr_norm1)
            UV_2D_fitting = feature_deviation(lrf_info.uv_ty, lrf_info.uv_y, self.uv_curve.max_y, self.uv_norm1)
        elif USE_JSD:
            XR_2D_fitting = deformed_jsd(lrf_info.xr_ty, lrf_info.y)
            UV_2D_fitting = deformed_jsd(lrf_info.uv_ty, lrf_info.uv_y)
        else:
            assert False, "invalid fitting method"
 
        XR_LRF_residual = np.log10(np.linalg.norm(self.W_*(Pxr @ Cxr - self.xrD_))/self.xr_norm2)
        UV_LRF_residual = np.log10(np.linalg.norm(Puv @ Cuv - mapped_UvD)/self.uv_norm2)

        Guinier_deviation = self.get_guinier_deviation(Pxr, Cxr, rg_params)
        SEC_conformance = self.compute_comformance(xr_params, rg_params, seccol_params)

        qv = self.gdev.qv
        kratky_plot_smoothness = 0
        for k, p in enumerate(Pxr.T[0:self.num_pure_components]):
            rg = self.gdev.rgs[k]
            if np.isnan(rg):
                continue
            iz = self.gdev.izs[k]
            kratky_y = (qv*rg)**2*p/iz

            # negative penalty
            kratky_plot_smoothness += np.mean(np.min([self.xr_zero_scatter, kratky_y], axis=0)**2)

            if self.kratky_smoothness:
                # smoothness
                # https://stackoverflow.com/questions/68023055/how-to-evaluate-the-smoothness-flatness-of-a-curve-in-python
                kratky_plot_smoothness += np.std(np.diff(kratky_y))

        if np.isnan(kratky_plot_smoothness):
            kratky_plot_smoothness = KRATKY_SMOOTHNESS_BAD_VALUE
        # basic_mean = np.mean([XR_2D_fitting, UV_2D_fitting, XR_LRF_residual, UV_LRF_residual])
        Kratky_smoothness = np.log10(kratky_plot_smoothness) - 0.6
        Guinier_deviation = Guinier_deviation

        if debug:
            self.logger.info("Guinier_deviation=%.3g, kratky_plot_smoothness=%.3g", Guinier_deviation, Kratky_smoothness)

        score_list = [XR_2D_fitting, XR_LRF_residual, UV_2D_fitting, UV_LRF_residual, Guinier_deviation, Kratky_smoothness, SEC_conformance]
        assert len(score_list) == self.NUM_MAJOR_SCORES

        if not COERCE_BOUNDED_BQ:
            # B(q) out of bounds
            if self.sf_bounds is not None:
                penalties[3] += self.sf_bounds.compute_penalty(Pxr)

        # negetive_penalty for P's
        for P in [Pxr, Puv]:
            P_ = P[:,:-1]   # excluding baseline component
            penalties[1] += WEAK_PENALTY_SCALE * np.linalg.norm(P_[P_ < 0])

        # common order_penalty
        valid_rg_params = rg_params[self.valid_components]
        if len(valid_rg_params) > 1:
            # Rg order_penalty
            penalties[4] += PENALTY_SCALE * np.sum(np.min([self.zeros_valid_rg, valid_rg_params[:-1] - valid_rg_params[1:]], axis=0)**2)

        if EVAL_PEAK_DEVIATION:
            penalties[4] += PENALTY_SCALE * self.compute_peak_deviation(xr_params, debug=debug)

        # to avoid cases as in 20230303/HasA where UV_LRF_residual is lowered
        # without considering the bad effect that it seriously sacrifices UV_2D_fitting
        control_penalty = WEAK_PENALTY_SCALE * ( 
                              max(0, XR_2D_fitting - XR_LRF_residual - SUPERIOR_2D_LRF_ALLOW)**2
                            + max(0, UV_2D_fitting - UV_LRF_residual - SUPERIOR_2D_LRF_ALLOW)**2
                            )
        penalties.append(control_penalty)

        discreteness = 0
        if self.apply_rg_discreteness:
            discreteness += compute_rg_discreteness(rg_params, unit=self.rg_discreteness_unit)

        if self.apply_mw_integrity:
            discreteness += self.compute_mw_integrity(Pxr, Cuv, UV_2D_fitting)

        if self.apply_discreteness:
            score_list.append(discreteness)

        score_array = np.array(score_list)
        fv = synthesize(score_array, positive_elevate=3) + np.sum(penalties)

        if np.isnan(fv):
            # NaN is not allowed in scipy.optimize.basinhopping
            if not self.isnan_logged:
                self.logger.info("fv is NaN: score_array=%s", str(score_array))
                self.isnan_logged = True
            fv = np.inf
        return fv, score_array

    def objective_func(self, p, plot=False, debug=False, fig_info=None, axis_info=None, return_full=False):
        # override this
        assert False

    def debug_plot_params(self, norm_params, **kwargs):
        plot = kwargs.pop("plot", True)
        return self.objective_func(self.to_real_params(norm_params), plot=plot, **kwargs)

    def objective_func_wrapper(self, norm_params, **kwargs):
        return self.objective_func(self.to_real_params(norm_params), **kwargs)

    def get_score_names(self, major_only=False):
        names = ([  "XR_2D_fitting", "XR_LRF_residual", "UV_2D_fitting", "UV_LRF_residual",
                    "Guinier_deviation", "Kratky_smoothness", "SEC_conformance",
                    "mapping_penalty", "negative_penalty", "baseline_penalty",
                    "outofbounds_penalty", "order_penalty", "control_penalty",
                    ]
                )
        if major_only:
            return names[0:self.NUM_MAJOR_SCORES]

        if self.apply_rg_discreteness or self.apply_mw_integrity:
            count = 0
            if self.apply_rg_discreteness:
                plotname = "discreteness: "
                rg_name = "%g" % self.rg_discreteness_unit
                count += 1
            else:
                rg_name = ""
            if self.apply_mw_integrity:
                plotname = "MW ratios: "    # this overwrite the above
                mw_name = "%s" % self.mw_integer_ratios
                count += 1
            else:
                mw_name = ""
            comma = ", " if count == 2 else ""
            names.insert(7, "%s%s%s%s" % (plotname, rg_name, comma, mw_name))
        return names

    def get_num_scores(self, penalties):
        n = self.NUM_MAJOR_SCORES + len(penalties)
        if self.apply_rg_discreteness:
            n += 1
        return n

    def get_num_matrices(self):
        return 5

    def set_params_scale(self, real_bounds, debug=False):
        lower = real_bounds[:,0]
        upper = real_bounds[:,1]
        self.scale_shift = lower
        self.scale_slope = (upper - lower)/PARAMS_SCALE
        size = len(real_bounds)
        self.xlower = np.zeros(size)
        self.xupper = np.ones(size)*PARAMS_SCALE

        if debug:
            with plt.Dp():
                fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
                for ax, (start, stop) in [
                                            (ax1, (0, -6)),
                                            (ax2, (-6, len(lower)))
                                            ]:
                    slice_ = slice(start, stop)
                    ax.set_title("set_params_scale %s" % (str((start, stop))))
                    ax.plot(self.scale_shift[slice_], label="lower(or scale_shift)")
                    ax.plot(upper[slice_], label="upper")
                    axt = ax.twinx()
                    axt.grid(False)
                    axt.plot(self.scale_slope[slice_], color="C2", label="scale_slope")
                    ax.legend()
                    axt.legend(loc="center left")
                fig.tight_layout()
                plt.show()

    def to_norm_params(self, real_params):
        conv_params = real_params.copy()   # for not to change the argument
        if self.use_K:
            # K = N*T
            # log(K) = log(N) + log(T)
            # p = log(T)/log(K)
            # m = me + mp
            # q = mp/m
            N, me, T, mp = real_params[self.nj:]
            K =  N*T
            m = me + mp
            conv_params[self.nj] = K
            conv_params[self.mej] = m
            conv_params[self.tj] = np.log(T)/np.log(K)
            conv_params[self.mpj] = mp/m
        nx = (conv_params - self.scale_shift)/self.scale_slope
        return nx

    def to_real_params(self, norm_params):
        rx = self.scale_slope*norm_params + self.scale_shift
        if self.use_K:
            K, m, p, q = rx[self.nj:]
            # N = K**(1 - p)
            # T = K**p
            # me = m*(1 - q)
            # mp = m*q
            rx[self.nj] = K**(1 - p)
            rx[self.mej] = m*(1 - q)
            rx[self.tj] = K**p
            rx[self.mpj] = m*q
        return rx

    def accept_test(self, **kwargs):
        x = kwargs["x_new"]
        tmax = np.all(x <= self.xupper)
        tmin = np.all(x >= self.xlower)
        return tmax and tmin

    def minima_callback(self, x, f, accept):
        real_params = self.to_real_params(x)
        self.logger.info("minima_callback: f=%.3g, accept=%s", f, str(accept))
        save_opt_params(self.cb_fh, real_params, f, accept, self.eval_counter)
        self.callback_counter += 1
        if self.shm is not None:
            self.shm.array[0] = self.callback_counter

        self.update_at_minima(x, f, accept)

        if IMMEDIATE_KNOWN_BEST:
            if self.min_fv is None or f < self.min_fv:
                self.min_fv = f
                self.min_i = self.fv_array_size
                self.fv_array[self.min_i] = f
                self.fv_array_size += 1
                self.update_bounds(real_params)

        return False

    def update_at_minima(self, x, f, accept):
        self.update_minima_props(x)

    def update_bounds(self, x, debug=False):
        if True:
            if self.exports_bounds:
                bounds_array = np.array(self.params_type.get_param_bounds(x, real_bounds=self.real_bounds))
            else:
                bounds_array = np.array(self.params_type.get_param_bounds(x))
            # self.logger.info("update_bounds(exports_bounds=%s): bounds_array=%s", self.exports_bounds, str(bounds_array))
            if bounds_array.shape[0] == len(x):
                pass
            else:
                error_msg = "bounds_array.shape[0](%d) != len(x)(%d)" % (bounds_array.shape[0], len(x))
                self.logger.error(error_msg)
                raise AssertionError(error_msg)

            self.lower_bounds = bounds_array[self.bounds_mask,0]
            self.upper_bounds = bounds_array[self.bounds_mask,1]
        else:
            scale = 2
            masked_init_params = x[self.bounds_mask]
            self.params_type.update_bounds_hook(masked_init_params)
            self.lower_bounds = masked_init_params / scale
            self.upper_bounds = masked_init_params * scale
        if debug:
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title("update_bounds")
                ax.plot(self.lower_bounds, label="lower_bounds")
                ax.plot(self.upper_bounds, label="upper_bounds")
                ax.legend()
                fig.tight_layout()
                plt.show()

    def update_bounds_hook(self, masked_init_params):
        # nothing to do
        pass

    def get_param_bounds(self, params):
        return self.params_type.get_param_bounds(params)

    def split_bounds(self, bounds):
        return self.params_type.split_bounds(bounds)

    def reshape_xr_bounds(self, xr_bounds):
        pass

    def xr_baseline(self, x, xr_baseparams, y_, cy_list):
        # note: update also ComplementaryView
        return self.xr_base_curve(x, xr_baseparams, y_, cy_list)

    def uv_baseline(self, x, uv_baseparams, y_, cy_list):
        # note: update also ComplementaryView
        return self.uv_base_curve(x, uv_baseparams, y_, cy_list)

    def get_name(self):
        import re
        name_re = re.compile(r"'([^']+)'")
        m = name_re.search(str(type(self)))
        name = m.group(1).split(".")[-1]
        return name

    def get_region_limit(self):
        # this should be properly overridden
        return 0

    def update_primary_peak(self, debug=False):
        xr_params = self.init_separate_params[0]
        if len(xr_params.shape) == 1:
            i = np.argmax(xr_params)
        else:
            i = np.argmax(xr_params[:,0])
        self.primary_component = i
        j = np.argmax(self.xr_curve.y)
        x = self.xr_curve.x
        self.xr_primary_x = x[j]
        self.xr_x_width = x[-1] - x[0]

        x = self.uv_curve.x
        y = self.uv_curve.y
        self.uv_x_width = x[-1] - x[0]
        a, b = self.init_separate_params[3]
        uv_x = a*self.xr_primary_x + b
        j = bisect_right(x, uv_x)
        start = max(0, j - 10)
        stop = min(len(x), j + 10)
        neighbor = slice(start, stop)
        m = np.argmax(y[neighbor])
        j_ = start + m
        self.uv_primary_x = x[j_]
        if debug:
            
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.plot(x, y)
                ax.plot(x[j_], y[j_], "o", color="red")
                fig.tight_layout()
                plt.show()

    def compute_peak_deviation(self, xr_params, debug=False):
        if self.use_K:
            # meaning if self.stochastic
            # currently there is no simple method to compute deviation for stochastic models
            return 0
        else:
            a, b = self.separate_params[3]
            xr_x = self.params_type.compute_xr_peak_position(self.primary_component, xr_params)
            xr_dev = (max(PRIMARY_PEAK_DEV_ALLOW, abs(xr_x - self.xr_primary_x)/self.xr_x_width) - PRIMARY_PEAK_DEV_ALLOW)**2
            uv_x = a*xr_x + b
            uv_dev = (max(PRIMARY_PEAK_DEV_ALLOW, abs(uv_x - self.uv_primary_x)/self.uv_x_width) - PRIMARY_PEAK_DEV_ALLOW)**2
            if debug:
                self.logger.info("xr_peak_deviation=%g, uv_peak_deviation=%g", xr_dev, uv_dev)
            return xr_dev + uv_dev

    def update_guinier_region(self, params=None, debug=False):
        if debug:
            from importlib import reload
            import molass_legacy.GuinierTools.GuinierDeviation
            reload(molass_legacy.GuinierTools.GuinierDeviation)
        from molass_legacy.GuinierTools.GuinierDeviation import GuinierDeviation

        if params is None:
            params = self.init_params
            rg_params = self.init_separate_params[2]
        else:
            rg_params = self.separate_params[2]     # this will be probably right, meaning that rg_params are consistant with params
        region_limit = self.get_region_limit()
        self.gdev = GuinierDeviation(self.qvector, self.rg_curve, rg_params, region_limit, composite=self.composite)
        
        lrf_info = self.objective_func(params, return_lrf_info=True)
        Pxr = lrf_info.matrices[0]
        Cuv = lrf_info.matrices[3]
        self.gdev.update_region(rg_params, region_limit, P=Pxr, M=self.xrD, E=self.xrE, debug=debug)

        if self.apply_mw_integrity:
            self.c_indeces = np.argmax(Cuv[:-1], axis=1)
            self.logger.info("updated c_indeces to %s", str(self.c_indeces))

    def get_guinier_deviation(self, P, Cxr, rg_params, debug=False):
        if USE_RGCURVE_DEVIATION:
            # use Optimizer.FvScoreConverter.compute_fv_adjustment to change the adjustment
            deviation = self.gdev.compute_rgcurve_deviation(Cxr, rg_params, adjust=self.rgcurve_adjust)
            return deviation 
        else:
            deviation = self.gdev.compute_deviation(P, Cxr, rg_params, valid_components=self.valid_components, debug=debug)
            if deviation is None:
                return
            return deviation*0.5 - 0.5

    def compute_comformance(self, xr_params, rg_params, seccol_params):
        conformance = self.params_type.compute_comformance(xr_params, rg_params, seccol_params, poresize_bounds=self.poresize_bounds)
        return conformance*0.5 - 0.1

    def compute_mw_integrity(self, P, C, preceder):
        if self.c_indeces is None:
            self.c_indeces = np.argmax(C[:-1], axis=1)
        cvector = np.array([C[i,j] for i,j in enumerate(self.c_indeces)])
        mw_ratios = self.gdev.izs / cvector
        if self.mw_integer_ratios is None:
            from .IntegerRatios import determine_integer_ratios
            self.mw_integer_ratios = determine_integer_ratios(mw_ratios)
        dist = compute_mw_integrity_impl(mw_ratios, self.mw_integer_ratios, preceder)
        return dist

    def compute_moments_list(self, params=None, debug=False):
        from molass_legacy.Peaks.MomentsUtils import compute_moments
        if params is None:
            params = self.init_params
        
        lrf_info = self.objective_func(params, return_lrf_info=True)
        x = lrf_info.x
        moments_list = []
        for cy in lrf_info.scaled_xr_cy_array[0:-1]:    # excluding baseline component
            M = compute_moments(x, cy)
            moments_list.append(M)
        if debug:
            y = lrf_info.y
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.plot(x, y, label="data")
                for k, (cy, M) in enumerate(zip(lrf_info.scaled_xr_cy_array, moments_list)):
                    ax.plot(x, cy, ":", label="component-%d" % k)
                    ax.axvline(M[1], color="green", label="M1=%g" % M[1])
                    s = np.sqrt(M[2])
                    ax.axvspan(M[1]-s, M[1]+s, color="green", alpha=0.2)
                ax.legend()
                fig.tight_layout()
                plt.show()
        return moments_list
    
    def get_strategy(self):
        from molass_legacy.Optimizer.Strategies.DefaultStrategy import DefaultStrategy
        return DefaultStrategy()
    
    def is_stochastic(self):
        return False


"""
    Theory.Rg.py

    Copyright (c) 2020-2021, SAXS Team, KEK-PF
"""
import numpy as np
from bisect import bisect_right
from scipy.stats import linregress
from molass_legacy.KekLib.ExceptionTracebacker import log_exception
from .SolidSphere import get_boundary_params_simple

GQ_LIMIT = 0.05

def compute_Rg(qv, I):
    x = qv**2
    y = np.log(I)
    slope, intercept = linregress(x, y)[0:2]
    Rg = np.sqrt(-3*slope)
    return Rg

def compute_corrected_Rg_impl(qv, M, E, initRg, gf, gt):
    x = qv[gf:gt]**2
    b1 = get_boundary_params_simple(initRg)[0]
    i = min(gt-1, bisect_right(qv, b1))
    c = M[i,:]
    C = np.array([c, c**2])
    P = M[0:gt,:] @ np.linalg.pinv(C)
    Aq = P[:,0]
    y = np.log(Aq[gf:])

    slope, intercept = linregress(x, y)[0:2]
    Rg = np.sqrt(-3*slope)

    if False:
        print('initRg=', initRg)
        print('Rg=', Rg)
        plt.push()
        fig, ax = plt.subplots()
        ax.plot(x, y)
        ax.plot(x, x*slope+intercept)
        fig.tight_layout()
        plt.show()
        plt.pop()

    return Rg

def compute_corrected_Rg(sd, ecurve, pno, qv, M, E, range_=None):
    if pno == ecurve.primary_peak_no:
        gf, gt_ = sd.pre_recog.get_gunier_interval()
        preRg = sd.pre_recog.get_rg()
    else:
        try:
            pre_rg = sd.pre_recog.pre_rg
            if range_ is None:
                # note that pno here can be buggy!
                selected = ecurve.peak_info[pno][1]
            else:
                selected = ecurve.get_peak_position(*range_)
            preRg = pre_rg.compute_rg(selected=selected)
            sg = pre_rg.sg
            gf = sg.guinier_start
            gt_ = sg.guinier_stop
        except:
            log_exception(None, "compute_corrected_Rg (1): ")

    print('qv[gt_]=', qv[gt_])
    if qv[gt_] > GQ_LIMIT:
        gt = bisect_right(qv, GQ_LIMIT)
        guinier_truncated = ' (truncated by q <= %g)' % GQ_LIMIT
    else:
        gt = gt_
        guinier_truncated = ''
    print((gf, gt), 'Rg=%.3g' % preRg)
    try:
        Rg = compute_corrected_Rg_impl(qv, M, E, preRg, gf, gt)
    except:
        # as in Kosugi8_Backsub
        log_exception(None, "compute_corrected_Rg (2): ")
        Rg = preRg

    return Rg, gf, gt, guinier_truncated



# coding: utf-8
"""
    OurProfiler.py

    Copyright (c) 2018-2019, Masatsuyo Takahashi, KEK-PF
"""
import time
import inspect
import cProfile, pstats, io

def profile(closure, site_packeges=False):
    pr = cProfile.Profile()
    pr.enable()

    closure()

    pr.disable()
    s = io.StringIO()
    sortby = 'cumulative'
    ps = pstats.Stats(pr, stream=s).sort_stats(sortby)
    ps.print_stats()

    frm = inspect.stack()[1]
    mod = inspect.getmodule(frm[0])
    file = mod.__file__ + '.prof'

    with open(file, "w") as fh:
        prof_str = s.getvalue()
        if site_packeges:
            fh.write(prof_str)
        else:
            for line in prof_str.split('\n'):
                if line.find('{') > 0 or line.find('<') > 0:
                    continue
                elif line.find('site-packages') > 0:
                    continue
                else:
                    fh.write( line + '\n' )

def take_time(closure, print_time=True):
    start = time.time()
    ret = closure()
    finish = time.time()
    t = finish - start
    if print_time:
        print( 'It took', t )
    return t, ret



"""
    AscDscDifference.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import molass_legacy.KekLib.DebugPlot as plt

def show_asc_dsc_difference(js_canvas, debug=True):
    print(show_asc_dsc_difference)

    if debug:
        with plt.Dp():
            fig, ax = plt.subplots()
            ax.plot(0, 0, "o")
            fig.tight_layout()
            plt.show()



"""
    LrfSolver.py

    Copyright (c) 2021-2025, SAXS Team, KEK-PF
"""
import logging
import numpy as np
from scipy.optimize import fmin_cg, minimize
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy._MOLASS.SerialSettings import get_setting
from .LrfInfoProxy import LrfInfoProxy

class LrfSolver:
    def __init__(self, pdata, popts, debug=False):
        if debug:
            from importlib import reload
            import molass_legacy.Extrapolation.ExtrapolationSolver
            reload(molass_legacy.Extrapolation.ExtrapolationSolver)   
        from molass_legacy.Extrapolation.ExtrapolationSolver import ExtrapolationSolver

        ExtrapolationSolver.__init__(self, pdata, popts)    # required in LrfResultPool
                                                            # for pno, paired_range in enumerate(solver.cnv_ranges)

        from importlib import reload
        import molass_legacy.LRF.PnoScdMap
        reload(molass_legacy.LRF.PnoScdMap)
        from .PnoScdMap import PnoScdMap

        self.xray_index = self.sd.xray_index
        self.logger = logging.getLogger(__name__)
        self.lrf_bound_correction = get_setting("lrf_bound_correction")
        self.logger.info("lrf_bound_correction=%d", self.lrf_bound_correction)
        self.impl = ExtrapolationSolver(pdata, popts)
        self.mc_vector = self.impl.mc_vector                # already scaled with conc_factor
        self.pno_map = PnoScdMap(pdata.sd, self.impl.cnv_ranges)

    def get_pno_map(self):
        return self.pno_map

    def solve_range(self, start, stop, peakset_info,
                    lrf_rank=1,
                    stop_event=None,
                    paired_range=None,
                    debug=False):

        self.logger.info("solve_range called with lrf_rank=%s", str(lrf_rank))

        done = False
        tried_bounded_lrf = False
        if self.lrf_bound_correction and lrf_rank >= 2:
            if debug:
                from importlib import reload
                import molass_legacy.BoundedLRF.BoundedLrfSolver
                reload(molass_legacy.BoundedLRF.BoundedLrfSolver)
                import molass_legacy.LRF.ElutionMatrix
                reload(molass_legacy.LRF.ElutionMatrix)
            from molass_legacy.BoundedLRF.BoundedLrfSolver import BoundedLrfSolver
            from .ElutionMatrix import ElutionMatrix

            tried_bounded_lrf = True
            try:
                em = ElutionMatrix(self.impl)
                C = em.make_elution_matrix(start, stop, peakset_info, lrf_rank=lrf_rank)
                if False:
                    np.savetxt("C-%d-%d.dat" % (start, stop), C)
                if len(C) >= 2:
                    C_ = C
                else:
                    pno, nth, peakset, known_peak_info = peakset_info
                    C_ = C[nth*2:(nth+1)*2]

                range_ = slice(start,stop)
                qv = self.qv
                D = self.data[:,range_]
                E = self.error[:,range_]

                i = self.xray_index
                j = paired_range.top_x
                Y = self.data[i,j]

                if debug:
                    debug_info = self.sd, range_
                else:
                    debug_info = None
                solver = BoundedLrfSolver(self.qv, D, E, C=C_, i=i, debug_info=debug_info)
                P_, C__, Rg, R_, L_, hK, hL, bq_bounds_, coerced_bq_ = solver.solve()
                aq_, bq_ = P_.T[0:2]

                Pe = solver.get_corrected_error()

                # scale = 1/self.impl.conc_factor
                top_c = self.mc_vector[j]
                Y_ = Y/top_c
                scale = Y_/aq_[i]

                Ae = Pe[:,0]*scale
                Be = Pe[:,1]*scale
                Z = None                # realy used?
                lrfE = [Ae, Be, Z]

                A = aq_*scale
                B = bq_*scale

                A_data = np.array([qv, A, Ae]).T
                B_ratio = np.max(B)/np.max(A)

                lrfinfo = LrfInfoProxy(A_data, need_bq=True, B_ratio=B_ratio, data=D, bq_bounds=(bq_bounds_[0]*scale, bq_bounds_[1]*scale))
                ret_C = C_              # is this backward compatible?

                if debug:
                    pass

                result = A, B, Z, lrfE, lrfinfo, ret_C
                done = True
            except:
                # task: identify this case for tests.
                from molass_legacy.KekLib.ExceptionTracebacker import warnlog_exception      # to avaid producing error messages
                warnlog_exception(self.logger, "BoundedLrfSolver failed: ")
                self.logger.info("resorting to the unbounded solver")

        if not done:
            if tried_bounded_lrf:
                self.logger.warning("bounded LRF failed. resorting to the traditional solver.")
            result =  self.impl.extrapolate_wiser(start, stop, peakset_info,
                                    stop_event=stop_event,
                                    lrf_rank=lrf_rank,
                                    )

        return result

class FrobeniusNorm:
    def __init__(self, M, C):
        self.M = M
        self.C = C
        self.p_shape = (M.shape[0], 2)

    def compute(self, x):
        P = x.reshape(self.p_shape)
        return np.linalg.norm(P @ self.C - self.M)**2

    def compute_gradient(self, x):
        P = x.reshape(self.p_shape)
        return ( 2 * ( (P @ self.C - self.M) @ self.C.T ) ).flatten()

class Regularization:
    def __init__(self, qv, R, L):
        self.qv = qv
        self.R = R
        self.L = L
        self.p_shape = (len(qv), 2)
        self.Kv = L**2/(qv*R)**4

    def compute(self, x):
        P = x.reshape(self.p_shape)
        p0 = P[:,0]
        p1 = P[:,1]
        dev = (p1/p0)**2 - self.Kv
        return np.sum(dev[dev > 0]**2)

    def compute_gradient(self, x):
        pass

def boundef_lrf(qv, M, C, Pinit, R, K=1):
    f = FrobeniusNorm(M, C)
    r = Regularization(qv, R, 1)
    Lam = 1

    def obj_func(x):
        return f.compute(x) + Lam*r.compute(x)

    def gradient_func(x):
        return f.compute_gradient(x) + Lam*r.compute_gradient(x)

    # ret = fmin_cg(obj_func, Pinit.flatten(), fprime=gradient_func)
    ret = minimize(obj_func, Pinit.flatten())
    return ret.x.reshape((len(qv),2))

def boundef_lrf_simple(qv, M, C, Pinit, R, K=1):
    p0 = Pinit[:,0].copy()
    p1 = Pinit[:,1].copy()
    L = 1
    bound = L/(qv*R)**2
    e = np.abs(p1/p0) - bound
    pos_e = e > 0
    d = (np.abs(p1[pos_e]) - bound[pos_e]*np.abs(p0[pos_e]))/(bound[pos_e]+1)
    p0 += d
    pospos_ep1 = np.logical_and(pos_e, p1 > 0)
    posneg_ep1 = np.logical_and(pos_e, p1 < 0)
    p1[pospos_ep1] -= (np.abs(p1[pospos_ep1]) - bound[pospos_ep1]*np.abs(p0[pospos_ep1]))/(bound[pospos_ep1]+1)
    p1[posneg_ep1] += (np.abs(p1[posneg_ep1]) - bound[posneg_ep1]*np.abs(p0[posneg_ep1]))/(bound[posneg_ep1]+1)
    return np.array([p0, p1]).T

def demo(root, sd, pno=0, debug=True):
    from bisect import bisect_right
    from molass_legacy.DataStructure.SvdDenoise import get_denoised_data
    from molass_legacy.Theory.Rg import compute_corrected_Rg
    from molass_legacy.Theory.SolidSphere import phi, get_boundary_params_simple
    from molass_legacy.Theory.SynthesizedLRF import synthesized_lrf_spike
    from molass_legacy.Theory.DjKinning1984 import S0
    from molass_legacy.Theory.SfBound import S0_bound

    M, E, qv, ecurve = sd.get_xr_data_separate_ly()

    range_ = ecurve.get_ranges_by_ratio(0.5)[pno]
    f = range_[0]
    p = range_[1]
    t = range_[2]

    pt = M[:,p]

    x = ecurve.x
    y = ecurve.y

    i = bisect_right(qv, 0.02)

    eslice = slice(f,t)
    M0 = M[:,eslice]
    Rg, gf, gt, _ = compute_corrected_Rg(sd, ecurve, pno, qv, M0, E[:,eslice])
    R = np.sqrt(5/3)*Rg

    norm_list = []

    c_ = y[eslice]

    M2 = get_denoised_data(M0, rank=2)
    C2 = np.array([c_, c_**2])
    P2 = M2 @ np.linalg.pinv(C2)

    norm_list.append(np.linalg.norm(P2@C2 - M0))

    b1, b2, k = get_boundary_params_simple(Rg)

    P_ = synthesized_lrf_spike(qv, M0, c_, M2, P2, boundary=b1, k=k)
    y1 = P_[:,0]
    y1n = y1/y1[0]

    Pb = boundef_lrf(qv, M2, C2, P_, R)

    fig, (ax0, ax1, ax2) = plt.subplots(ncols=3, figsize=(22,7))
    fig.suptitle(r"$\frac{L}{(qR)^2}$ Bounded LRF Demo", fontsize=30)
    ax0.set_title("I(q) = P(q)*S(q)", fontsize=16)
    ax1.set_title("P(q): Form Factors", fontsize=16)
    ax2.set_title("S(q): Structure Factors", fontsize=16)
    for ax in [ax0, ax1]:
        ax.set_yscale('log')

    py0 = phi(qv, R)**2
    sy0 = S0(qv, R*5.4/2, K=1)

    b1i = bisect_right(qv, b1)
    iy0 = py0*sy0
    ptyn = pt*iy0[b1i]/pt[b1i]

    ax0.plot(qv, iy0, color='green')
    ax0.plot(qv, ptyn)

    ax1.plot(qv, py0, color='green')
    ax1.plot(qv, y1n)
    ax1.plot(qv, Pb[:,0])

    s1y = 1 + P_[:,1]/y1
    ax2.plot(qv, sy0, color='green')
    ax2.plot(qv, s1y)
    ax2.plot(qv, 1 + Pb[:,1]/Pb[:,0])

    for j, ax in enumerate([ax0, ax1, ax2]):
        ymin, ymax = ax.get_ylim()
        ymax_ = 1.5 if j == 2 else ymax
        ax.set_ylim(ymin, ymax_)
        for b in [b1, b2]:
            ax.plot([b, b], [ymin, ymax_], ':', color='gray')

    R_ = R*5.4
    sbu = S0_bound(qv, R_, K=-2)
    sbl = S0_bound(qv, R_, K=2)
    ax2.plot(qv, sbu, ':', color='red')
    ax2.plot(qv, sbl, ':', color='red')


    fig.tight_layout()
    fig.subplots_adjust(top=0.8)
    plt.show()



"""

    NonAsciiPaths.py

    Copyright (c) 2023, SAXS Team, KEK-PF

"""

def are_all_single_bytes(s):
    return len(s.encode()) == len(s)

def nonascii_path_check(path, parent):
    if are_all_single_bytes(path):
        return True

    if not parent.env_info.atsas_is_available:
        # no need to ask
        return True

    parent.set_an_folder_error()
    parent.update()

    import molass_legacy.KekLib.CustomMessageBox as MessageBox
    yes = MessageBox.askyesno("Multibyte Path Option Query",
        '"%s"\n'
        "seems to include multi-byte characters\n"
        "which may not be processed properly in ATSAS programs.\n"
        "You can simply continue without ATSAS programs,\n"
        "or retry after replacing it with another single-byte only character folder.\n"
        "Woud you like to continue simply?" % path,
        parent=parent,
        )
    if yes:
        from Env.EnvInfo import EnvInfo, set_global_env_info
        from molass_legacy._MOLASS.SerialSettings import set_setting
        set_setting('revoke_atsas', 1)
        env_info = EnvInfo()
        env_info.atsas_is_available = False
        parent.tmp_logger.info("ATSAS has been revoked.")
        parent.env_info = env_info
        parent.analyzer.env_info = env_info     # this is problematic. consider better implementation
        set_global_env_info(env_info)
        parent.an_folder_entry.config(fg='black')
        parent.update()
        return True
    else:
        return False

if __name__ == '__main__':
    for path in [r"D:\田中\Data", r"D:\Data"]:
        print(path, are_all_single_bytes(path))



"""
    Stochastic.StationaryMove.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
from shapely import LineString, Point, Polygon, intersection
from molass_legacy.KekLib.PolygonGeometry import makeOffsetPoly
from molass_legacy.KekLib.CircleGeometry import circle_line_segment_intersection

REFLECT_LIMIT_DIST = 1e-5
REFLECT_NUM_TRIALS = 5

def find_nearest_wall(grain, px, py):
    cx, cy = grain.center
    dx = px - cx
    dy = py - cy
    angle = np.arctan2(dy, dx)
    entries = grain.entries.flatten()
    k = np.argmin((entries - angle)**2)
    i, j = divmod(k, 2)
    return i, j

def compute_boundary_walls(cx, cy, R, r, entry_points):
    outerx = [cx]
    outery = [cy]
    for ex, ey in entry_points:
        outerx.append(ex)
        outery.append(ey)
    innerx, innery = makeOffsetPoly(outerx, outery, -r, outer_ccw=1)
    ipoints = np.array([(ex, ey) for ex, ey in zip(innerx, innery)])
    cpoint = ipoints[0,:]
    bpoints = [cpoint]
    for ipoint in ipoints[1:,:]:
        dvec = ipoint - cpoint
        dvec /= np.linalg.norm(dvec)
        opoint = cpoint + dvec * R
        intersections = circle_line_segment_intersection((cx, cy), R, cpoint, opoint, full_line=False)
        if len(intersections) > 0:
            bpoints.append(intersections[0])
        else:
            # investigate this case
            pass
    return bpoints

def mirrorImage(a, b, c, x1, y1):
    """
    borrowed from
    https://www.geeksforgeeks.org/find-mirror-image-point-2-d-plane/
    """
    temp = -2 * (a * x1 + b * y1 + c) /(a * a + b * b)
    x = temp * a + x1
    y = temp * b + y1 
    return (x, y)

def compute_reflected_point(bpoints, wx, wy, nx, ny, debug=False):
    ls = LineString([(wx, wy), (nx, ny)])
    cpoint = bpoints[0]
    ret = None
    for point in bpoints[1:]:
        tp = intersection(ls, LineString([cpoint, point]))
        if debug:
            print(tp, type(tp) is Point)
        if type(tp) is Point:
            # task: choose the closer wall if both walls have an intersection point
            cx, cy = cpoint
            px, py = tp.x, tp.y
            """
            m = (py - cy)/(px - cx)
            y - cy = m * (x - cx)
            m*x - y + cy - m*cx = 0
            """
            m = (py - cy)/(px - cx)
            a = m
            b = -1
            c = cy - m*cx
            mx, my = mirrorImage(a, b, c, nx, ny)
            dist = np.sqrt((mx - nx)**2 + (my - ny)**2)
            if dist > REFLECT_LIMIT_DIST:
                if debug:
                    print("dist=", dist)
                    print("(nx, ny) ==> (mx, my)", (nx, ny), (mx, my))
                ret = mx, my
    return ret

def get_next_position_impl(particle, grain, last_px, last_py, px, py, debug_info=None):
    inmobile = False
    cx, cy = grain.center
    nx, ny = px, py
    i, j = find_nearest_wall(grain, px, py)
    entry_points = [grain.get_point_from_angle(angle) for angle in grain.entries[i,:]]
    bpoints = compute_boundary_walls(cx, cy, grain.radius, particle.radius, entry_points)
    wx, wy = last_px, last_py
    debug = debug_info is not None
    reflected = False
    for k in range(REFLECT_NUM_TRIALS):
        ret = compute_reflected_point(bpoints, wx, wy, nx, ny, debug=debug)
        if ret is None:
            break
        wx, wy = nx, ny
        nx, ny = ret
        reflected = True

    if not reflected:
        poly = Polygon(bpoints)
        if not poly.contains(Point(nx, ny)):
            intersections = circle_line_segment_intersection((cx, cy), grain.radius, (wx, wy), (nx, ny), full_line=False)
            if len(intersections) == 0:
                dist = np.sqrt((nx - cx)**2 + (ny - cy)**2)
                if dist < grain.radius + particle.radius:
                    inmobile = True
            else:
                if debug:
                    print("not reflected", intersections[0])
                tx, ty = intersections[0]
                """
                get the tangent line equation
                and find the mirror image of (nx, ny) against the tangent line
                ((x, y) - (tx, ty)) ・ ((cx, cy) - (tx, ty)) = 0
                (x - tx, y - ty)・(cx - tx, cy - ty) = 0
                (x - tx, y - ty)・(a, b) = 0
                a*x + b*y - a*tx - b*ty = 0
                """
                a = cx - tx
                b = cy - ty
                c = - a*tx - b*ty
                nx, ny = mirrorImage(a, b, c, nx, ny)
                reflected = True
                inmobile = True

    if debug_info is not None:
        from matplotlib.patches import Polygon as MplPolygon
        print("get_next_position_impl", px, py)
        print("get_next_position_impl", nx, ny, inmobile)
        fig, ax = debug_info
        ax.cla()
        ax.set_title("get_next_position_impl debug")  
        grain.draw(ax)
        dx = px - last_px
        dy = py - last_py
        vlen = np.sqrt(dx**2 + dy**2)
        ax.arrow(x=last_px, y=last_py, dx=dx, dy=dy, width=0.0005, head_width=0.002, length_includes_head=True,
                    head_length=0.2*vlen, color='black', alpha=0.5)
        for entry in entry_points:
            ax.plot([cx, entry[0]], [cy, entry[1]], color='red')

        if reflected:
            dx = nx - wx
            dy = ny - wy
            vlen = np.sqrt(dx**2 + dy**2)
            ax.arrow(x=wx, y=wy, dx=dx, dy=dy, width=0.0005, head_width=0.002, length_includes_head=True,
                    head_length=0.2*vlen, color='green', alpha=0.5)
            
        poly = MplPolygon(bpoints, color="red", alpha=0.5)
        ax.add_patch(poly)
        particle.draw(ax, alpha=0.5)

        r = grain.radius * 1.2
        ax.set_xlim(cx - r, cx + r)
        ax.set_ylim(cy - r, cy + r)
        fig.canvas.draw_idle()

    return nx, ny, inmobile


# coding: utf-8
"""
    SecTheory.StochasticModels.py

    Copyright (c) 2022, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.special import iv
from scipy.optimize import basinhopping
import molass_legacy.KekLib.DebugPlot as plt

def stochastic(x, h, tpi, npi, scale, t0):
    x_ = x*scale - t0
    return h * iv(1, np.sqrt(4*npi*x_/tpi)) * np.sqrt(npi/(x_*tpi)) * np.exp(-x_/tpi-npi)

def demo(parent, in_folder, trimming=True, correction=False, p0=None):
    from scipy.optimize import curve_fit
    from molass_legacy._MOLASS.SerialSettings import set_setting
    from molass_legacy.Batch.StandardProcedure import StandardProcedure
    from molass_legacy.Baseline.BaselineUtils import get_corrected_sd_impl
    from molass_legacy.Trimming.PreliminaryRecognition import PreliminaryRecognition
    from molass_legacy.Elution.CurveUtils import simple_plot
    from molass_legacy.Peaks.ElutionModels import egh
    from LPM import get_corrected
    from DataUtils import get_in_folder

    sp = StandardProcedure()
    sd = sp.load_old_way(in_folder)

    pre_recog = PreliminaryRecognition(sd)
    if trimming:
        sd_ = sd._get_analysis_copy_impl(pre_recog)
    else:
        sd_ = sd.get_copy()

    if correction:
        v2_copy = get_corrected_sd_impl(sd_, sd, pre_recog)
    else:
        v2_copy = sd_

    D, _, wv, ecurve = v2_copy.get_uv_data_separate_ly()

    ex = ecurve.x[10:-10]
    ey = get_corrected(ecurve.y[10:-10])

    p_init = ecurve.get_emg_peaks()[ecurve.primary_peak_no].get_params()
    popt1, pcov1 = curve_fit(egh, ex, ey, p_init)

    if p0 is None:
        # p0 = (400, 25, 25, 1, 0)
        # p0 = (274.72277122, 18.89459478, 49.19590951, 1.3624587, 13.62451978)
        p0 = (275.1072381, 14.57463207, 63.54627583, 1.35989494, 13.59206267)

    if True:
        with plt.Dp():
            fig, ax = plt.subplots()
            ax.set_title("init debug")
            ax.plot(ex, ey)
            ax.plot(ex, stochastic(ex, *p0))
            plt.show()
    if True:
        def obj_func(p):
            y = stochastic(ex, *p)
            ret = np.sum((y - ey)**2)
            if np.isfinite(ret):
                return ret
            else:
                return 1e10
        for k in range(1):
            print([k], "---- basinhopping")
            result = basinhopping(obj_func, p0, stepsize=10)
            p0 = result.x
        popt2 = result.x
    else:
        popt2, pcov2 = curve_fit(stochastic, ex, ey, p0)

    print("popt2=", popt2)
    with plt.Dp():
        in_folder = get_in_folder(in_folder)
        fig, ax = plt.subplots()
        ax.set_title("Stochastic Model Trial for %s" % in_folder, fontsize=20)
        # simple_plot(ax, ecurve, legend=False)
        ax.plot(ex, ey, label="data")
        ax.plot(ex, egh(ex, *popt1), ":", lw=2, label="egh")
        ax.plot(ex, stochastic(ex, *popt2), ":", lw=2, label="stochastic model")
        ax.legend(fontsize=16)
        fig.tight_layout()
        plt.show()



"""
    Optimizer.ResultProxy.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import os
import logging
import numpy as np
from .BasicOptimizer import BasicOptimizer
from .FullOptResult import FullOptResult
from .OptJobInfo import OptJobInfo
from .FuncImporter import import_objective_function

class ResultProxy(FullOptResult):
    def __init__(self, folder, first_result):
        self.logger = logging.getLogger(__name__)
        self.folder = folder
        self.first_result = first_result
        self.optimizer = None
        self.first_optimizer = first_result.get_optimizer()
        self.dsets = first_result.dsets
        info = OptJobInfo()
        info.load(self.folder)
        self.n_components = info.nc

    def get_optimizer(self):
        if self.optimizer is None:
            composite = self.load_composite()
            log_file = self.get_log_file()
            class_code = self.get_class_code(log_file)
            optimizer_class = import_objective_function(class_code, self.logger)[0]
            first_optimizer = self.first_optimizer
            self.optimizer = optimizer_class(self.dsets, self.n_components,
                                uv_base_curve=first_optimizer.uv_base_curve,
                                xr_base_curve=first_optimizer.xr_base_curve,
                                qvector=first_optimizer.qvector,
                                wvector=first_optimizer.wvector,
                                composite=composite,
                                )
    
            self.fv_list, self.x_list, self.fv_array= self.load_callback_info()

            if class_code == 'G1100':
                from Estimators.SdmEstimatorProxy import SdmEstimatorProxy
                estimator = SdmEstimatorProxy(self.folder)
                self.optimizer.params_type.set_estimator(estimator)      # reconsider the neccesity of this line

            self.optimizer.debug_fv = False
            init_params = self.load_init_params()
            self.optimizer.prepare_for_optimization(init_params, minimally=True)

        return self.optimizer


"""
    Optimizer.Strategies.BasicStrategy.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""

class BasicStrategy:
    def __init__(self, cycle=None):
        self.cycle = cycle

    def trust_initial_baseline(self):
        return True

    def baseline_first(self):
        return True

    def is_strategic(self, n):
        if self.cycle is None:
            return False
        if n % self.cycle == 1:
            return True
        return False


"""
    G1100.py

    Copyright (c) 2024-2025, SAXS Team, KEK-PF
"""
import numpy as np
from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker
from molass_legacy.Models.Stochastic.DispersivePdf import dispersive_monopore_pdf as elutionmodel_func
from molass_legacy.Optimizer.BasicOptimizer import BasicOptimizer, PENALTY_SCALE, UV_XR_RATIO_ALLOW, UV_XR_RATIO_SCALE
from molass_legacy.Optimizer.NumericalUtils import safe_ratios
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy.ModelParams.SeccolFunctions import rgfit_secconf_eval
from molass_legacy.Optimizer.TheDebugUtils import convert_score_list
from molass_legacy.Optimizer.PenaltyUtils import compute_mapping_penalty

EGH_LOG_ALPHA = np.log(0.1)
TAU_BOUND_RATIO = get_setting("TAU_BOUND_RATIO")    # tau <= sigma*TAU_BOUND_RATIO
LRF_RESIDUAL_FAKED = 10
XR_VALID = 0.001
RG_FITTING_NAN_REPLACE = 100

class G1100(BasicOptimizer):
    """
    Stochastic Dispersive Model
    """
    def __init__(self, dsets, n_components, **kwargs):
        self.elutionmodel_func = elutionmodel_func
        if True:
            from importlib import reload
            import molass_legacy.ModelParams.SdmParams
            reload(molass_legacy.ModelParams.SdmParams)
        from molass_legacy.ModelParams.SdmParams import SdmParams

        params_type = SdmParams(n_components)
        BasicOptimizer.__init__(self, dsets, n_components, params_type, kwargs)
        self.exports_bounds = True

    def objective_func(self, p, plot=False, debug=False, fig_info=None, axis_info=None, return_full=False, avoid_pinv=False, return_lrf_info=False, **kwargs):
        self.eval_counter += 1
        xr_params, xr_baseparams, rg_params, (a, b), uv_params, uv_baseparams, (c, d), sdmcol_params = self.split_params_simple(p)

        x = self.xr_curve.x
        y = self.xr_curve.y
        # rg = self.rg

        N, K, x0, poresize, N0, tI = sdmcol_params
        me = 1.5
        mp = 1.5
        T = K/N
        rho = rg_params/poresize
        rho[rho > 1] = 1
        ty = np.zeros(len(x))

        uv_x = a*x+b
        uv_y = self.uv_curve.spline(uv_x)

        mapping_penalty = compute_mapping_penalty(self.uv_curve, self.xr_curve, self.init_mapping, (a, b), len(self.uv_curve.x), xr_params, uv_params)

        masked_params = p[self.bounds_mask]
        outofbounds_penalty = PENALTY_SCALE * (np.sum(np.max([self.zero_bounds, self.lower_bounds - masked_params], axis=0)) + np.sum(np.max([self.zero_bounds, masked_params - self.upper_bounds], axis=0)))
        if self.eval_counter == 1 and outofbounds_penalty > 0:
            self.logger.info("out of lower bounds: %s", str(np.max([self.zero_bounds, self.lower_bounds - masked_params], axis=0)))
            self.logger.info("out of upper bounds: %s", str(np.max([self.zero_bounds, masked_params - self.upper_bounds], axis=0)))
        if plot:
            overlap = np.zeros(len(x))
            overlap_penalities = []

        xr_cy_list = []
        uv_cy_list = []
        xr_ty = np.zeros(len(x))
        uv_ty = np.zeros(len(uv_x))
        negative_penalty = min(0, T)**2
        T_ = abs(T)
        k = 0
        x_ = x - tI
        t0 = x0 - tI
        for xr_w, r_, uv_w in zip(xr_params, rho, uv_params):
            negative_penalty += min(0, xr_w)**2 + min(0, uv_w)**2
            np_ = N*(1 - r_)**me
            tp_ = T_*(1 - r_)**mp
            pd_cy = elutionmodel_func(x_, np_, tp_, N0, t0)
            xr_cy = xr_w * pd_cy
            uv_cy = uv_w * pd_cy

            xr_ty += xr_cy
            xr_cy_list.append(xr_cy)
            uv_ty += uv_cy
            uv_cy_list.append(uv_cy)
            k += 1

        xr_cy = self.xr_baseline(x, xr_baseparams, xr_ty, xr_cy_list)
        uv_cy = self.uv_baseline(uv_x, uv_baseparams, uv_ty, uv_cy_list)
        xr_ty += xr_cy
        xr_cy_list.append(xr_cy)
        uv_ty += uv_cy
        uv_cy_list.append(uv_cy)

        try:
            lrf_info = self.compute_LRF_matrices(x, y, xr_cy_list, xr_ty, uv_x, uv_y, uv_cy_list, uv_ty, debug=debug)
            if return_lrf_info:
                return lrf_info

            m, s = xr_baseparams[0:2]   # how about r?

            slope_penalty = max(self.slope_allowance, (m - self.init_slope)**2) - self.slope_allowance
            intercept_penalty = max(self.intercept_allowance, (s - self.init_intercept)**2) - self.intercept_allowance
            baseline_penalty = slope_penalty*self.slope_penalty_scale + intercept_penalty*self.intercept_penalt_scale
            negative_penalty = PENALTY_SCALE * (negative_penalty + intercept_penalty)
            order_penalty = 0       # common order_penalty will be added in compute_fv

            penalties = [mapping_penalty, negative_penalty, baseline_penalty, outofbounds_penalty, order_penalty]

            fv, score_list = self.compute_fv(lrf_info, xr_params, rg_params, sdmcol_params, penalties, p, debug=debug)
        except:
            etb = ExceptionTracebacker()
            last_lines = etb.last_lines(n=2)
            if last_lines.find("SVD") > 0:
                if self.svd_error_count == 0:
                    self.logger.warning( "error in objective_func: " + last_lines)
                self.svd_error_count += 1
                svd_error = True
            else:
                self.logger.warning( "error in objective_func: " + last_lines)
                svd_error = False
            lrf_info = self.create_lrf_info_for_debug(x, y, xr_ty, xr_cy_list, uv_x, uv_y, uv_ty, uv_cy_list)
            if return_lrf_info:
                return lrf_info
            fv = np.inf
            penalties = [0] * 6     # above penalties + [control_penalty]
            score_list = [0] * self.get_num_scores([])      # score_list does not include penalties here

            if svd_error and not avoid_pinv and debug:
                self.objective_func(p, plot=True, fig_info=fig_info, axis_info=axis_info, avoid_pinv=True)
                return fv

        if plot:
            from importlib import reload
            import molass_legacy.ModelParams.SdmPlotUtils
            reload(molass_legacy.ModelParams.SdmPlotUtils)
            from molass_legacy.ModelParams.SdmPlotUtils import plot_objective_state

            print("fv=", fv)

            debug_fv = plot_objective_state((score_list, penalties), fv, None,
                lrf_info,
                overlap, self.rg_curve, rg_params,
                self.get_score_names(),
                fig_info, axis_info,
                self, p,
                avoid_pinv=avoid_pinv,
                **kwargs
                )
            if axis_info is None:
                self.debug_fv = debug_fv

        if return_full:
            score_list = convert_score_list((score_list, penalties))
            return fv, score_list, *lrf_info.matrices
        else:
            return fv

    def get_strategy(self):
        from molass_legacy.Optimizer.Strategies.SdmStrategy import SdmStrategy
        return SdmStrategy(nc=self.n_components - 1)
    
    def is_stochastic(self):
        return True


"""

"""

import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..')))

from molass_legacy.KekLib.ExcelCOM import CoInitialize, CoUninitialize, ExcelComClient

def test_excel_com():
    CoInitialize()
    xlValue = 2
    save = True
    book_path = os.path.join( os.path.dirname(__file__), 'test_excel_com.xlsx' )
    exce_client = ExcelComClient(visible=True)
    ws = exce_client.openWorksheet( book_path + '(TestSheet)' )
    line1 = ws.draw_line( ( 100, 50 ), ( 100, 210 ) )
    line2 = ws.draw_line( ( 150, 50 ), ( 150, 210 ) )
    brace = ws.draw_right_brace( ( 110, 220 ), 20, 50, 90 )

    for i, shape in enumerate( [ line1, line2, brace ] ):
        shape.Select()
        line_ = exce_client.selection().ShapeRange.Line
        line_.ForeColor.RGB =   0x50B000    # RGB(0, 176, 80)
        line_.Weight = 2 if i < 2 else 1

    ws.draw_textbox( "テスト文字列", ( 110, 250 ), 100, 20, visible=False )
    ws.freeze_panes( "C2" )

    if False:
        # assert ws.get_num_charts() == 1

        chart = ws.get_chart( 0 )
        print( 'chart=', chart )
        chart.Axes(xlValue).MaximumScale = 70
        area = chart.ChartArea
        X, Y = area.Left, area.Top
        point = ws.get_chart_data_point( chart, 0, 4 )
        x, y = point.Left, point.Top
        ws.draw_oval( (X+x, Y+y), 10, 10, color=0x0000FF )

    if save:
        ws.workbook.Save()
        ws.workbook.Close()     # necessary for other processes to access

    exce_client.quit()
    CoUninitialize()


"""
    VersionUtils.py

    Copyright (c) 2024, Masatsuyo Takahashi, KEK-PF
"""
import re
import os
from subprocess import run
from datetime import datetime

def get_exe_version(exe_path):
    ret = run([exe_path, '-V'], capture_output=True)
    return [int(v) for v in ret.stdout.decode()[:-1].split(" ")[-1].split('.')]

def update_molass_version_file(version_file, with_time=False):
    dist_folder = version_file.split('\\')[-4]
    version  = dist_folder.split('-')[-2].replace('_', '.')
    version_re = re.compile(r"_MOLASS .+ python")
    temp_file = version_file + ".temp"
    out_fh = open(temp_file, "w")
    with open(version_file) as in_fh:
        for line in in_fh:
            if line.find('_MOLASS') > 0:
                format = "%Y-%m-%d"
                if with_time:
                    format += " %H:%M:%S"
                line = re.sub(version_re, "_MOLASS %s (%s python" % (version, datetime.today().strftime(format)), line)
            out_fh.write(line)
    out_fh.close()
    os.remove(version_file)
    os.rename(temp_file, version_file)

if __name__ == '__main__':
    import sys
    print(get_exe_version(sys.executable))


"""
    EnvMessages.py

    Copyright (c) 2019-2022, SAXS Team, KEK-PF
"""
from .EnvInfo import get_global_env_info

class LogPicker:
    def __init__(self):
        self.log_info = []

    def info(self, m):
        self.log_info.append([0, m])

    def warning(self, m):
        self.log_info.append([1, m])

def get_env_messages():
    messages = []

    picker = LogPicker()
    envinfo = get_global_env_info(gpu_info=True)
    envinfo.show_and_log_if_unavailable(None, picker)
    for level, message in picker.log_info:
        messages.append(message + "\n")

    messages.append("Reason:\n")

    texts = envinfo.get_gpu_reason_texts()
    for text in texts:
        messages.append("    " + text + "\n")

    return messages



# coding: utf-8
"""
    HeadAnomalies.py

    Copyright (c) 2017, Masatsuyo Takahashi, KEK-PF
"""

import numpy        as np
from scipy          import stats

HEAD_SIZE               = 10
NEXT_SIZE               = HEAD_SIZE * 3
ACCEPTABLE_DEVIATION    = 1.0
NUM_CONSECUTIVE_ACCEPTS = 2

DEBUG = False

class HeadAnomalies:
    def __init__( self, y, X, Y ):
        self.X  = X
        self.Y  = Y

    def remove_head_anomalies( self ):
        try:
            self.remove_head_anomalies_impl()
        except:
            # this case will be caused e.g. when X, X are empty
            self.start_point = 0

    def remove_head_anomalies_impl( self ):
        start_candidate = np.argmax( self.Y )
        stop_candidate  = len(self.Y)//2
        if DEBUG: print( 'start_candidate=', start_candidate, len(self.Y) )

        last_accept = None
        num_accepts = 0
        for i in range( start_candidate, stop_candidate ):
            next_start  = i + HEAD_SIZE
            head_slope  = self.evaluate_slope( i, next_start )
            next_slope  = self.evaluate_slope( next_start, next_start + NEXT_SIZE )
            deviation   = abs( head_slope/next_slope - 1 )
            if DEBUG: print( i, 'deviation=', deviation )
            if deviation <= ACCEPTABLE_DEVIATION:
                if last_accept is not None and i == last_accept + 1:
                    num_accepts += 1
                    if num_accepts == NUM_CONSECUTIVE_ACCEPTS:
                        start_candidate = i - num_accepts + 1
                        break
                else:
                    num_accepts = 1
                last_accept = i

        self.start_point = start_candidate

    def evaluate_slope( self, start, stop ):
        x, y = self.X[start:stop], self.Y[start:stop]
        slope, intercept, r_value, p_value, std_err = stats.linregress( x, y )
        return slope



"""

    EmgPeak.py

    Copyright (c) 2018-2024, SAXS Team, KEK-PF

"""
import numpy as np
from scipy.interpolate import UnivariateSpline
import logging
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.KekLib.BasicUtils import ordinal_str
from molass_legacy.KekLib.NumpyUtils import moving_average
from molass_legacy.KekLib.OurMatplotlib import get_default_colors
from molass_legacy.Models.ElutionCurveModels import EGH, egh_x_from_height_ratio
from molass_legacy.KekLib.ExceptionTracebacker import log_exception, warnlog_exception
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy.Elution.CurveUtils import simple_plot

DEVIATION_LIMIT_RATIO   = 0.01
GLOBAL_ADOPTTION_RATIO  = 1.1
FIT_LIMITS_MIN_RATIO    = 0.05  # > 0.029 and < 0.070 for Ald;
                                # > 0.035 and < 0.074 for OA_Ald
                                # > 0.023 and < 0.183 for Open01
LOCAL_IMPROVE_RATIO     = 0.1
WIDEN_ALLOWANCE_RATIO   = 0.1
EXTENDABLE_DEPRESSION_RATIO = 0.3   # < 0.333 for BL-10C/Ald
WIDE_RANGE_RATIO        = 0.02
USE_MODEL_TOP_X_RANGES  = False
EMG_PEAK_SKIP_OK_RATIO  = 0.1
SIGMA_POINT_RATIO = 2

if USE_MODEL_TOP_X_RANGES:
    from ScipyUtils import erfcxinv

def proof_plot( peak, x_, y_, x, y, fy0, fy1, ry, fx, tx, fy2, chisqrs, chisqrs_, opt_label ):
    from matplotlib.ticker import FormatStrFormatter
    import matplotlib.ticker    as ticker

    # FuncFormatter can be used as a decorator
    @ticker.FuncFormatter
    def major_formatter(x, pos):
        # print( 'x, pos=', x, pos )
        if pos is not None:
            if  x - int(x) == 0:
                return "%.3g" % x
            else:
                return ""
        else:
            return "%g" % x

    colors = get_default_colors()

    plt.push()
    fig = plt.figure( figsize=(16, 6) )
    ax1 = fig.add_subplot( 121 )
    ax2 = fig.add_subplot( 122 )
    ax1.set_title( "Fitted model lines and the final residual line", fontsize=16 )
    ax2.set_title( "Chisquares (Σr²/n) of the fits", fontsize=16 )

    ax1.set_xlabel( "Elution No.", fontsize=16 )
    ax1.set_ylabel( "Intensity", fontsize=16 )
    ax2.set_xlabel( "Fit Trial Order", fontsize=16 )
    ax2.set_ylabel( "Chisquare", fontsize=16 )

    ax2.yaxis.set_major_formatter(FormatStrFormatter('%.3g'))
    ax2.xaxis.set_major_formatter(major_formatter)

    ax1.plot( x_, y_, color='orange' )
    ax1.plot( peak.top_x, peak.top_y, 'o', color='red' )
    for k, fy in enumerate( [ fy0, fy1, fy2 ] ):
        ax1.plot( x, fy, ':', label=ordinal_str(k) + ' fit' )

    ax1.plot( x, ry, ':', label='final %s residual' % opt_label )

    for i in [fx, tx]:
        ax1.plot( i, y_[i-x_[0]], 'o', color='cyan' )

    hy  = peak.top_y / 2
    ax1.plot( [ peak.top_x - peak.hwhmL, peak.top_x + peak.hwhmR ], [ hy, hy ], color='red' )
    ax1.legend(fontsize=12)

    gl_texts = [ 'global', 'local' ]

    for j, cs in enumerate( [chisqrs, chisqrs_] ):
        fmt = '-' if j == 0 else ':'
        ax2.plot( cs, fmt, color='gray', label=gl_texts[j] )
        for k, c in enumerate(cs):
            ax2.plot( k, c, 'o', label='%s-chisqr%d' % (gl_texts[j], k), color=colors[k] )

    ax2.legend(fontsize=12)

    fig.tight_layout()
    plt.show()
    plt.pop()

class EmgPeak:
    def __init__(self, x_size, top_x, top_y, area_prop=None):
        self.logger = logging.getLogger(__name__)
        self.x_size = x_size
        self.top_x  = int( top_x + 0.5 )
        self.top_y  = top_y
        self.area_prop = area_prop      # this will be set later in UnifiedDecompResult

    def shift_copy(self, x0, x_stop):
        # note that this copy is not complete.
        # refactoring desired
        # also, removing lmfit package should be considered
        from copy import deepcopy

        top_x = self.top_x + x0
        epeak_copy = EmgPeak(x_stop, top_x, self.top_y)
        epeak_copy.model = self.model
        epeak_copy.opt_params = deepcopy(self.opt_params)
        try:
            epeak_copy.opt_params[1] += x0
        except:
            assert False
        half_width = SIGMA_POINT_RATIO * epeak_copy.opt_params[2]
        epeak_copy.sigma_points = [ top_x - half_width, top_x + half_width ]

        return epeak_copy

    def estimate_params( self, x_, y_, lower, upper, allow_wider=False, debug=False ):
        widen_allow = int( len(x_) * WIDEN_ALLOWANCE_RATIO )

        slice_ = slice( lower, upper+1 )
        x   = x_[slice_]
        y   = y_[slice_]

        hm  = self.top_y/2

        if debug:
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title("estimate_params entry")
                ax.plot( x_, y_ )
                ax.plot( x, y )
                ax.plot( self.top_x, self.top_y, 'o', color='red' )
                ax.plot( x[[0,-1]], [hm, hm], color='red' )
                plt.show()

        try:
            indL = np.where( np.logical_and( x < self.top_x, y > hm ) )[0][0]
            indR = np.where( np.logical_and( x > self.top_x, y > hm ) )[0][-1]
        except:
            raise Exception( 'failed to find indL or indR' )

        # print( 'indL, indR=', indL, indR, x[indL], x[indR] )

        self.indL   = indL
        self.indR   = indR

        self.hwhmL  = self.top_x - x[indL]
        self.hwhmR  = x[indR] - self.top_x
        self.fwhm   = self.hwhmL + self.hwhmR

        model = EGH()   # EGH is the most robust among EGH, EMG, and EMGA.

        def get_wider_xy():
            wf  = max(0, lower-widen_allow)
            wt  = min(len(x_)-1, upper+widen_allow)
            slice_ = slice( wf, wt+1 )
            wx  = x_[slice_]
            wy  = y_[slice_]
            return wx, wy

        eval_interval = [ indL, indR ]
        f, t = 0, len(x)-1
        fy0, chisqr0, chisqr0_, params0  = self.try_fit( model, x, y, f, t, eval_interval, 0, debug=debug)
        fy1, chisqr1, chisqr1_, params1  = self.try_fit( model, x, y, indL, indR, eval_interval, 1, debug=debug )
        opt_got = False
        init_global_ratio   = chisqr0/chisqr0_
        local_improve_ratio = chisqr1_/chisqr0_
        # print( 'init_global_ratio=', init_global_ratio, 'local_improve_ratio=', local_improve_ratio )
        # local_improve_ratio > LOCAL_IMPROVE_RATIO is neccesary for the 1st peak of Feb25
        if init_global_ratio < GLOBAL_ADOPTTION_RATIO and local_improve_ratio > LOCAL_IMPROVE_RATIO  or allow_wider:
            opt_label = '0th'
            wx, wy = get_wider_xy()
            f, t = 0, len(wx)-1
            offset = max(0, lower-widen_allow)
            eval_interval_w = [ _ + offset for _ in eval_interval ]

            chisqrw = None
            for k, wmodel in enumerate([model]):
                try:
                    fy_, chisqrw, chisqrw_, paramsw = self.try_fit( wmodel, wx, wy, f, t, eval_interval_w, 1.5, debug=debug )
                    if debug:
                        print( '------------------------------------- chisqrw=', chisqrw, 'chisqr0=', chisqr0 )
                    opt_got = True
                    break
                except AssertionError as exc:
                    if debug:
                        log_exception(self.logger, "EmgPeak.estimate_params: ")
                    continue

            if chisqrw is not None and chisqrw < chisqr0:
                opt_f_x, opt_t_x = wx[0], wx[-1]
                opt_params = paramsw
                if debug:
                    self.logger.info("opt_params have been normally set from paramsw with mu=%g", opt_params["tR"].value)
            else:
                opt_f_x, opt_t_x = x[0], x[-1]
                opt_params = params1    # ?

        if debug or not opt_got:
            ry  = moving_average( np.abs( fy1 - y ), keepsize=True )
            f, t = self.find_fit_limits( x, y, fy0, fy1, ry, debug=debug )
            fy2, chisqr2, chisqr2_, params2 = self.try_fit( model, x, y, f, t, eval_interval, 2, debug=debug )

            if opt_got:
                ry  = moving_average( np.abs( fy0 - y ), keepsize=True )
            else:
                # this path is necessary for Ald, etc.
                opt_f_x, opt_t_x = x[f], x[t]
                opt_label = '2nd'
                opt_params = params2
                self.logger.warning("opt_params have been covered from paramsw with mu=%g", opt_params[1])

        if debug:
            wx, wy = get_wider_xy()
            proof_plot( self, wx, wy, x, y, fy0, fy1, ry, opt_f_x, opt_t_x, fy2, [ chisqr0, chisqr1, chisqr2 ], [chisqr0_, chisqr1_, chisqr2_ ], opt_label )

        self.flimL  = opt_f_x
        self.flimR  = opt_t_x
        self.model = model
        self.opt_params = opt_params
        self.sigma_points = self.get_sigma_points( SIGMA_POINT_RATIO )
        # assert self.sigma_points[0] < self.top_x_m and self.top_x_m < self.sigma_points[1]

    def get_model_y(self, x):
        return self.model.eval( self.opt_params, x=x )

    def get_model_x_from_ratio(self, ratio):
        params = self.opt_params
        mu = params[1]
        sigma = params[2]
        tau = params[3]
        ret = egh_x_from_height_ratio(ratio, mu, sigma, tau)
        return ret

    def get_sigma_points(self, ratio, max_sigma_tau=False):
        # TODO: verify or refine this
        params = self.opt_params

        try:
            a = params[4]
        except:
            a = 0

        sigma   = params[2]
        tau = params[3]

        if USE_MODEL_TOP_X_RANGES:
            try:
                mu  = params[1]                     # EMG
                if abs(tau) > 0.1:
                    # this doesn't yet seem to be correct
                    model_top_x = mu + erfcxinv(tau/sigma*np.sqrt(2/np.pi))*sigma*np.sqrt(2) - sigma**2/tau + a
                else:
                    model_top_x = mu + a
            except Exception as exc:
                # print(exc)
                raise exc
        else:
            model_top_x = self.top_x

        if max_sigma_tau:
            scale = max(sigma, tau)
        else:
            scale = sigma

        f = int(max(0, model_top_x - scale*ratio))
        t = int(min(self.x_size-1, model_top_x + scale*ratio))

        self.top_x_m = int(model_top_x + 0.5)
        return [ f, t ]

    def __repr__( self ):
        return 'EmgPeak' + str( ( self.flimL, self.top_x, self.flimR ) )

    def get_fit_limits( self ):
        return self.flimL, self.flimR

    def try_fit( self, model, x, y, f, t, eval_interval, did, debug=False ):
        slice_ = slice( f, t+1 )
        x_  = x[slice_]
        y_  = y[slice_]
        params = model.guess( y_, x=x_ )

        fy = None
        for method in ['least_squares']:
            try:
                out = model.fit(y_, params, x=x_, method=method)
                fy  = model.eval( out.params, x=x )
                break
            except:
                # numpy.linalg.LinAlgError: SVD did not converge, etc.
                warnlog_exception(self.logger, "method=%s failed in model.fit" % method)
                continue

        assert fy is not None

        chisqr  = np.sum( (fy - y)**2 ) / len(x)
        eval_slice = slice( eval_interval[0], eval_interval[1]+1 )
        eval_fy = fy[eval_slice]
        eval_y  = y[eval_slice]
        chisqr_ = np.sum( ( eval_fy - eval_y )**2 ) / len(eval_y)

        if debug:
            plt.push()
            # RuntimeWarning: invalid value encountered in double_scalars
            fig, ax = plt.subplots()
            ax.set_title( "try_fit %g" % did )
            ax.plot( x, y )
            ax.plot( x, fy )
            for k in [f, t]:
                ax.plot( x[k], y[k], 'o', color='cyan' )
            plt.show()
            plt.pop()

        return fy, chisqr, chisqr_, out.params

    def find_fit_limits( self, x, y, fy0, fy1, ry, debug=False ):
        # TODO: consider cases with bad S/N
        delta = self.top_y* DEVIATION_LIMIT_RATIO
        w   = np.where( np.logical_and( x < self.top_x, ry > delta ) )[0]
        f   = w[-1] if len(w) > 0 else 0
        w   = np.where( np.logical_and( x > self.top_x, ry > delta ) )[0]
        t   = w[0] if len(w) > 0 else len(x) - 1

        if debug:
            plt.push()
            ax = plt.gca()
            ax.cla()
            ax.set_title("find_fit_limits")
            ax.plot( x, y )
            for k in [f, t]:
                ax.plot( x[k], y[k], 'o', color='yellow' )
            plt.show()
            plt.pop()

        fx  = x[0:self.indL]
        fy  = ry[0:self.indL]
        try:
            if True:
                f_ratio = np.max(fy) / self.top_y
                # print( 'f_ratio=', f_ratio )
                if f_ratio < FIT_LIMITS_MIN_RATIO:
                    # TODO: this test is not stable
                    f   = 0
            else:
                spline = UnivariateSpline( fx, fy )
                fd0 = spline(fx)
                fd1 = spline.derivative(1)(fx)
                fd2 = spline.derivative(2)(fx)
                len_neg = len( np.where( fd2 < 0 )[0] )
                f_ratio = len_neg / len(fx)
                # print( 'f_ratio=', f_ratio, len_neg, len(fx) )
                if f_ratio < EXTENDABLE_DEPRESSION_RATIO:
                    f   = 0
        except:
            fd2 = None

        tx  = x[self.indR+1:]
        ty  = ry[self.indR+1:]
        try:
            if True:
                t_ratio = np.max(ty) / self.top_y
                # print( 't_ratio=', t_ratio )
                if t_ratio < FIT_LIMITS_MIN_RATIO:
                    # TODO: this test is not stable
                    t   = len(x) - 1
            else:
                spline = UnivariateSpline( tx, ty )
                td0 = spline(tx)
                td1 = spline.derivative(1)(tx)
                td2 = spline.derivative(2)(tx)
                len_neg = len( np.where( td2 < 0 )[0] ) 
                t_ratio = len_neg / len(tx)
                # print( 't_ratio=', t_ratio, len_neg, len(tx) )
                if t_ratio < EXTENDABLE_DEPRESSION_RATIO:
                    t   = len(x) - 1
        except:
            td2 = None

        f   = int( min( self.indL, f) )
        t   = int( max( self.indR, t) )

        if debug:
            # self.extend_debug_plot( x, y, fy0, fy1, ry, f, t, fx, fd0, fd1, fd2, tx, td0, td1, td2 )
            self.extend_debug_plot( x, y, fy0, fy1, ry, f, t, fx, tx )

        return f, t

    # def extend_debug_plot( self, x, y, fy0, fy1, ry, f, t, fx, fd0, fd1, fd2, tx, td0, td1, td2 ):
    def extend_debug_plot( self, x, y, fy0, fy1, ry, f, t, fx, tx ):
        plt.push()
        ax  = plt.gca()
        ax.cla()
        ax_ = ax.twinx()
        ax_.cla()
        ax.plot( x, y )
        ax.plot( x, fy0, ':', label='fy0' )
        ax.plot( x, fy1, ':', label='fy1' )
        ax.plot( x, ry )
        ax.plot( self.top_x, self.top_y, 'o', color='red' )
        hm  = self.top_y/2
        ax.plot( x[ [self.indL, self.indR] ], [hm, hm], color='red' )

        for i in [f, t]:
            ax.plot( x[i], y[i], 'o', color='orange' )

        if False:
            labels = [ 'fd0', 'fd1', 'fd2', 'td0', 'td1', 'td2' ]
            for k, point in enumerate( [ [fx, fd0], [fx, fd1], [fx, fd2], [tx, td0], [tx, td1], [tx, td2] ] ):
                dx, dy = point
                if dy is None:
                    continue

                if k in [0, 3]:
                    ax.plot( dx, dy, ':', label=labels[k] )
                else:
                    ax_.plot( dx, dy, ':', label=labels[k] )
                    neg_dy = dy < 0
                    ax_.plot( dx[neg_dy], dy[neg_dy], ':', color='red' )

        ax.legend( loc='upper left' )
        ax_.legend( loc='upper right' )
        plt.show()
        plt.pop()

    def get_assert_info(self):
        return ( self.flimL, self.top_x, self.flimR )

    def get_xkey( self ):
        return '%.3g' % (self.top_x)

    def get_params(self):
        return self.opt_params[0:4]

class EmgPeakLite(EmgPeak):
    def __init__( self, L, M, R):
        self.flimL  = L
        self.top_x  = M
        self.flimR  = R

class EmgPeakProxy(EmgPeak):
    def __init__( self, top_x, top_y, area_prop=None, model=None, opt_params=None):
        self.top_x  = top_x
        self.top_y  = top_y
        self.area_prop = area_prop
        self.model = model
        self.opt_params = opt_params
        sigma = opt_params[2]
        self.flimL = int(top_x - sigma)     # ok?
        self.flimR = int(top_x + sigma)     # ok?
        self.sigma_points = [ top_x - SIGMA_POINT_RATIO*sigma, top_x + SIGMA_POINT_RATIO*sigma ]
        self.sign = 1

def get_peaks( curve, max_y=None, allow_wider=False, orig_y=None, logger=None, debug=False ):

    local_debug = get_setting("local_debug")

    x   = curve.x
    y   = curve.y

    # if debug or len(curve.peak_info) == 0:
    if debug or local_debug:
        import inspect
        for frm in inspect.stack()[1:10]:
            print("get_peaks: %s %s (%d)" % (frm.filename, frm.function, frm.lineno))

        print("(1) curve.peak_info=", curve.peak_info)

        with plt.Dp():
            fig, ax = plt.subplots()
            ax.set_title("get_peaks entry")
            simple_plot(ax, curve)
            plt.show()

    if False:
        try:
            wide_ranges = curve.get_ranges_by_ratio(WIDE_RANGE_RATIO)
            # print( 'wide_ranges=', wide_ranges )
        except:
            log_exception(logger, "curve.get_ranges_by_ratio")

    peaks = []
    for k, info in enumerate(curve.peak_info):
        lower, top, upper = [x[j] for j in info]
        ratio = (top - lower)/(upper - lower)
        if debug or local_debug:
            print([k], (lower, top, upper), ratio)
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title("%d-th peak plot" % k)
                ax.plot(x, y)
                for px, label in [(lower, 'lower'), (top, 'top'), (upper, 'upper')]:
                    ax.plot(px, curve.spline(px), "o", label=label)
                ax.legend()
                fig.tight_layout()
                plt.show()

        if ratio < EMG_PEAK_SKIP_OK_RATIO:
            # for cases such as the 2nd peak of 20181203
            if logger is not None:
                from molass_legacy.KekLib.BasicUtils import ordinal_str
                logger.warning( '%s peak has been skipped due to ratio=%.3g in get_peaks' % (ordinal_str(k+1), ratio) )
            continue
        if top < lower:
            # fix this bug for the residual curve of May11
            continue

        if max_y is not None:
            if curve.spline(top) < max_y:
                continue

        peak = EmgPeak( len(x), top, curve.spline( top ) )
        try:
            peak.estimate_params( x, y, lower - x[0], upper - x[0], allow_wider=allow_wider, debug=debug)
            if orig_y is None:
                peak.sign = 1
            else:
                peak.sign = 1 if orig_y[peak.top_x] > 0 else -1

            peaks.append( peak )
        except:
            if logger is not None:
                warnlog_exception(logger, "get_peaks: ", n=5)
            continue

        if False:
            plt.push()
            fig, ax = plt.subplots()
            ax.set_title("get_peaks: %dth peak debug" % k)
            ax.plot(x, y, label="data")
            y_ = peak.get_model_y(x)
            ax.plot(x, y_, ':', color="C%d" % (k+1), label="%dth peak" % k)
            ptx = info[1]
            ax.plot(ptx, curve.spline(ptx), "o", color="red")
            ax.legend()
            fig.tight_layout()
            plt.show()
            plt.pop()

    if debug and len(peaks) == 0:
        # this can be a case for pH6
        print( 'peak_info=', curve.peak_info )
        plt.push()
        fig, ax = plt.subplots()
        ax.set_title("get_peaks: No emgpeak!")
        simple_plot(ax, curve)
        fig.tight_layout()
        plt.show()
        plt.pop()
        # assert False

    return peaks



"""
    Models.RateTheory.EdmBasicStudy.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np

def study():
    import seaborn as sns
    import molass_legacy.KekLib.DebugPlot as plt
    from molass_legacy.Models.RateTheory.RobustEDM import edm_impl

    sns.set_theme()

    t0 = 50
    u = 1
    a = 1.0
    # b = -4.0
    b = 1.0
    e = 0.3
    Dz = 0.2
    cinj = 1

    x = np.arange(500)
   
    with plt.Dp():
        fig, ax = plt.subplots()
        ax.set_title("Varying e: Total Porosity", fontsize=16)
        for e in [0.5, 0.4, 0.3, 0.2, 0.1]:
            y = edm_impl(x, t0, u, a, b, e, Dz, cinj)
            ax.plot(x, y, label="e=%g" % e)
        ax.legend()
        fig.tight_layout()
        plt.show()

    e = 0.2

    with plt.Dp():
        fig, ax = plt.subplots()
        ax.set_title("Varying cinj: Concentrationn at Injection", fontsize=16)
        for cinj in [1, 2, 3, 4, 5]:
            y = edm_impl(x, t0, u, a, b, e, Dz, cinj)
            ax.plot(x, y, label="cinj=%g" % cinj)
        ax.legend()
        fig.tight_layout()
        plt.show()
        
    cinj = 1

    with plt.Dp():
        fig, ax = plt.subplots()
        ax.set_title("Varying Dz: Dispersion Coefficient", fontsize=16)
        for Dz in [0.2, 0.3, 0.5, 0.7, 0.9]:
            y = edm_impl(x, t0, u, a, b, e, Dz, cinj)
            ax.plot(x, y, label="Dz=%g" % Dz)
        ax.legend()
        fig.tight_layout()
        plt.show()

    cinj = 1
    Dz = 0.5

    with plt.Dp():
        fig, ax = plt.subplots()
        ax.set_title("Varying b: Isotherm Parameter", fontsize=16)
        for b in [-1, -0.5, 0.001, 0.5, 1]:
            y = edm_impl(x, t0, u, a, b, e, Dz, cinj)
            ax.plot(x, y, label="b=%g" % b)
        ax.legend()
        fig.tight_layout()
        plt.show()

if __name__ == "__main__":
    import sys
    sys.path.append("..\\lib")
    study()


"""
    PeakSetSelector.py

    Copyright (c) 2018-2024, SAXS Team, KEK-PF
"""
import logging
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.Mapping.SingleComponent import PEAK_EVAL_RANGE_RATIO
from molass_legacy.KekLib.ExceptionTracebacker import log_exception

class KnownRec:
    def __init__(self, rno, pno, k_info):
        self.rno = rno
        self.pno = pno
        self.k_info = k_info

    def __repr__(self):
        return '<KnownRec(rno=%d, pno=%d, k_info=%s)>' % (self.rno, self.pno, str(self.k_info))

def get_range_from_elm_recs(p, num, k, f, t, elm_recs, ecurve=None, debug=False):
    x = ecurve.x

    if debug:
        y = ecurve.y
        plt.push()
        fig, ax = plt.subplots()
        ax.set_title("get_range_from_elm_recs debug: [%d-%d] %d:%d" % (p, k, f, t))
        ax.plot(x, y, color='orange')
        axt = ax.twinx()

    hmax = None
    top = (f+t)//2
    for i, rec in enumerate(elm_recs):
        evaluator = rec[1]
        # h = evaluator.get_param_value(0)
        """
        h does not necessarily indicate the hight
        as observed in 20190529_4
        """
        h = evaluator(x[top])

        if debug:
            print([i], 'h=', h)
            c = rec[0]
            y = evaluator(x)
            axt.plot(x, y, color='C%d'%c, label=str(c))

        if hmax is None or h > hmax:
            try:
                f, t = evaluator.x_from_height_ratio(ecurve, PEAK_EVAL_RANGE_RATIO)
                hmax = h
                if debug:
                    print([i], "set: (f, t)=", (f, t), "h=", h)
            except:
                if False:
                    with plt.Dp():
                        print([i], "debug plot")
                        y_ = ecurve.y
                        fig, ax = plt.subplots()
                        ax.set_title("get_range_from_elm_recs failure debug")
                        ax.plot(x, y_)
                        ax.plot(x, evaluator(x))
                        fig.tight_layout()
                        plt.show()
                log_exception(None, "get_range_from_elm_recs failure: ")

    if debug:
        ymin, ymax = ax.get_ylim()
        ax.set_ylim(ymin, ymax)
        for p in [f, t]:
            p_ = int(round(p))
            ax.plot([p_,p_], [ymin, ymax], ':', color='red', label=str(p_))
        ax.legend(loc='upper left')
        axt.legend(loc='upper right')
        plt.show()
        plt.pop;()

    return int(round(f)), int(round(t))

class PeakSetSelector:
    def __init__(self, paired_ranges, ecurve=None, debug=False):
        self.logger = logging.getLogger(__name__)
        self.paired_ranges = paired_ranges
        x0 = ecurve.x[0]    # x0 == 0 for traditional v1 usage

        if debug:
            x = ecurve.x
            y = ecurve.y
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title("PeakSetSelector entry")
                ax.plot(x, y)
                for p, paired_range in enumerate(paired_ranges):
                    for k, rec in enumerate(paired_range.elm_recs):
                        func = rec[1]
                        ax.plot(x, func(x), ":", label="%d-%d" % (p,k))
                ax.legend()
                fig.tight_layout()
                plt.show()

        range_indeces = []
        pfoot_indeces = []
        for p, paired_range in enumerate(paired_ranges):
            elm_recs = paired_range.elm_recs
            fromto_list = paired_range.get_fromto_list()
            num_fromto = len(fromto_list)
            for k, (f, t) in enumerate(fromto_list):
                range_indeces.append( [p, f + x0, t + x0])
                if elm_recs is None:
                    pfoot_indeces.append( [p, f + x0, t + x0])
                else:
                    f_, t_ = get_range_from_elm_recs(p, num_fromto, k, f, t, elm_recs, ecurve=ecurve, debug=debug)
                    pfoot_indeces.append( [p, f_, t_])

        self.logger.info("range_indeces have been created as follows")
        for n, rec in enumerate(range_indeces):
            self.logger.info("[%d] %s", n, str(rec))
        self.logger.info("pfoot_indeces have been created as follows")
        for n, rec in enumerate(pfoot_indeces):
            self.logger.info("[%d] %s", n, str(rec))

        self.required_peakset_infos = []
        self.known_peak_info_list = None
        self.row_info_dict = {}
        """
            e.g. required_peakset_infos is constructed as follows for OA_Ald

            [row] [peak0, n-th-in-the-selected, {peak0, peak1, ...} ]

            [0] [0, 0, {0}]         # row [0] represents peak0 which is 0-th in {peak0}
            [1] [0, 0, {0, 1}]      # row [1] represents peak0 which is 0-th in {peak0, peak1}
            [2] [1, 1, {0, 1}]      # row [2] represents peak1 which is 1-th in {peak0, peak1}
            [3] [2, 0, {2}]         # row [3] represents peak2 which is 0-th in {peak2}
            [4] [2, 0, {2}]         # row [4] represents peak2 which is 0-th in {peak2}

            note that numbers in the required_peakset_infos represent peaks not elements.
            (although in this case, "peaks" and "elements" are identical,
            there are cases where a single peak consists of more than one elements)

            for a microfluidic paried range, it should be

            [0] [0, 0, {0, 1}]
            [1] [1, 1, {0, 1}]
        """

        """
            f ---------------- t
                  f_ --------------------- t_
             max(f,f_)  <=  min(t,t_)
               means there is some overlap
        """
        self.logger.info("required_peakset_infos are created as follows")
        row = 0
        for p, f, t in range_indeces:
            peaks = set()
            for p_, f_, t_ in pfoot_indeces:
                max_f = max(f, f_)
                min_t = min(t, t_)
                if max_f <= min_t:
                    peaks.add( p_ )
            try:
                nth = list(peaks).index(p)
            except:
                # verify that this should never occur
                nth = 0
                self.logger.warning("to cope with this unexpected situation, nth is set to 0")

            self.required_peakset_infos.append( [p, nth, peaks] )
            self.logger.info("[%d] %s", row, str([p, nth, peaks]))
            row_info = self.row_info_dict.get(p)
            if row_info is None:
                row_info = self.row_info_dict[p] = []
            row_info.append(row)
            row += 1

        if debug:
            from matplotlib.patches import Rectangle

            print('PeakSetSelector ---- debug info ---')
            print('paired_ranges=', paired_ranges)
            print('range_indeces=', range_indeces)
            print('pfoot_indeces=', pfoot_indeces)
            print('required_peakset_infos=', self.required_peakset_infos)

            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title("PeakSetSelector Debug for" + str(paired_ranges))
                x = ecurve.x
                y = ecurve.y
                ax.plot(x, y, color='orange')
                ymin, ymax = ax.get_ylim()
                ax.set_ylim(ymin, ymax)
                def plot_ranges(ymin_, ymax_, indeces, color, label_fmt):
                    for pk, f, t in indeces:
                        p = Rectangle(
                                (f, ymin_),     # (x,y)
                                t - f,          # width
                                ymax_ - ymin_,  # height
                                facecolor = color,
                                alpha = 0.2,
                                label = label_fmt % pk,
                            )
                        ax.add_patch(p)
                plot_ranges(ymin, ymax, range_indeces, 'cyan', "%dth range")
                plot_ranges(ymin, ymax/2, pfoot_indeces, 'pink', "%dth foot")
                ax.legend()
                plt.show()

    def update_known_peak_info_list(self, known_info_list):
        # print('update_known_peak_info_list: required_peakset_infos=', self.required_peakset_infos)
        # print('update_known_peak_info_list: known_info_list=', known_info_list)
        self.known_peak_info_list = []
        for row, info in enumerate(self.required_peakset_infos):
            pno, nth, peakset = info

            known_peak_info = []
            for ref_pno in peakset:
                known_pair_list = []
                for r in self.row_info_dict[ref_pno]:
                    known_pair_list.append( (abs(r - row), known_info_list[r]) )
                peak_info = sorted(known_pair_list)[0][1]
                known_peak_info.append(peak_info)
            self.known_peak_info_list.append(known_peak_info)

    def select_peakset(self, row):
        peakset_main_info = self.required_peakset_infos[row]
        if self.known_peak_info_list is None:
            known_peak_info = None
        else:
            known_peak_info = self.known_peak_info_list[row]
        self.logger.info("selected %d-th range has %s element(s).", row, len(peakset_main_info[2]))
        return  peakset_main_info + [known_peak_info]

    def select_demo_ranges_for_gd(self, uv_y, opt_recs):
        print('paired_ranges=', self.paired_ranges)
        print('required_peakset_infos=', self.required_peakset_infos)

        max_top_y = None
        max_top_x = None
        max_pno = None
        for p, paired_range in enumerate(self.paired_ranges):
            top_x = paired_range.top_x
            print([p], top_x)
            top_y = uv_y[top_x]
            if max_top_y is None or top_y > max_top_y:
                max_top_y = top_y
                max_top_x = top_x
                max_pno = p

        selected_pno = None
        for pno, peakset_info in enumerate(self.required_peakset_infos):
            if max_pno == peakset_info[0]:
                selected_pno = pno
                break

        print('max_pno=', max_pno, 'max_top_x=', max_top_x)
        selected_eno = None
        for eno, opt_rec in enumerate(opt_recs):
            top_x = opt_rec.peak.top_x
            if abs(top_x - max_top_x) < ALLOW_WIDER:
                selected_eno = eno
                break

        return self.required_peakset_infos[selected_pno]+[None], selected_eno
